// Init.cpp : Application initialisation and setup.
//

#include "stdafx.h"

#define USE_DIRECTDRAW    TRUE

/*
 * Global settings... must be initialised by user app in user_init.
 */

unsigned int viewresolution_x, viewresolution_y;
BOOL view_fullscreen;

/*
 * Global variables attached to this source file
 */
BOOL exit_application = 0;
BOOL requires_shutdown = 0;
HINSTANCE app_hinstance = 0;
HWND app_hwnd = 0;
HDC   app_hdc = 0;
HGLRC app_hglrc = 0;
FILE *debug_file = NULL;
DWORD last_frame_displayed = 0;

LPDIRECTDRAW        g_DirectDraw = NULL;
LPDIRECTINPUT       g_DirectInput = NULL;
LPDIRECTINPUTDEVICE g_DIKeyboard = NULL;
LPDIRECTINPUTDEVICE g_DIMouse = NULL;


// Register the window class.
int init_window_class(HINSTANCE hInstance)
{
  WNDCLASS wc;

  wc.style         = 0;
  wc.lpfnWndProc   = (WNDPROC) MainWndProc;
  wc.cbClsExtra    = 0;
  wc.cbWndExtra    = 0;
  wc.hInstance     = hInstance;
  wc.hIcon         = LoadIcon (hInstance, APP_NAME);
  wc.hCursor       = LoadCursor( NULL, IDC_ARROW );
  wc.hbrBackground = NULL;
  wc.lpszMenuName  = APP_NAME;
  wc.lpszClassName = APP_NAME;

  if (!RegisterClass(&wc)) {
    MessageBox(0, "Couldn't register window class", APP_NAME, MB_ICONERROR);
    return 0;
  }

  return TRUE;
}

// Create, show and focus the application's main window.
int init_main_window(HINSTANCE hInstance)
{
  app_hwnd = CreateWindow(APP_NAME, APP_NAME, WS_POPUP | WS_CLIPSIBLINGS | WS_CLIPCHILDREN,
                          0, 0, viewresolution_x, viewresolution_y,
                          NULL, NULL, hInstance, NULL);

  if (!app_hwnd) {
    MessageBox( 0, "Couldn't create window", APP_NAME, MB_ICONERROR);
    return 0;
  }

  ShowWindow(app_hwnd, SW_SHOWNORMAL);
	UpdateWindow(app_hwnd);
	SetFocus(app_hwnd);

  return TRUE;
}

// Change display modes so that we can run fullscreen as desired.
int init_display_mode(void)
{
#if USE_DIRECTDRAW
  HRESULT hr;

  hr = DirectDrawCreate(NULL, &g_DirectDraw, NULL);
  if (FAILED(hr)) {
    MessageBox(0, "Can't obtain DirectDraw interface", APP_NAME, MB_ICONERROR);
    return 0;
  }

  // Set to exclusive mode, so that we can change the color depth.
  hr = g_DirectDraw->SetCooperativeLevel(app_hwnd, DDSCL_EXCLUSIVE | DDSCL_FULLSCREEN);
  if (FAILED(hr)) {
    MessageBox(0, "Can't enter DirectDraw Exclusive Mode", APP_NAME, MB_ICONERROR);
    g_DirectDraw->Release();
    return 0;
  }

  // Change the display mode to the desired resolution, 16-bit color.
  hr = g_DirectDraw->SetDisplayMode(viewresolution_x, viewresolution_y, 16);
  if (FAILED(hr)) {
    char errbuf[256];
    sprintf(errbuf, "Can't change to resolution to %dx%d 16-bit",
            viewresolution_x, viewresolution_y);
    MessageBox(0, errbuf, APP_NAME, MB_ICONERROR);
    g_DirectDraw->Release();
    return 0;
  }

  // Exit exclusive mode, so that the OpenGL driver will work.
  hr = g_DirectDraw->SetCooperativeLevel(app_hwnd, DDSCL_NORMAL);
  if (FAILED(hr)) {
    MessageBox(0, "Can't revert to DirectDraw Normal Cooperative Level", APP_NAME, MB_ICONERROR);
    g_DirectDraw->Release();
    return 0;
  }

  return 1;
#else
  DEVMODE devmode;

  devmode.dmBitsPerPel=0;
  for (int i = 0; i < MAX_ENUMS; i++) {
    if (!EnumDisplaySettings(NULL, i, &devmode)) {
      MessageBox(0, "640x480 fullscreen 16-bit not supported", APP_NAME, MB_ICONERROR);
      return 0;
    } else if ((devmode.dmBitsPerPel == 16) &&
               (devmode.dmPelsWidth  == viewresolution_x) &&
               (devmode.dmPelsHeight == viewresolution_y)) {
      break;
    }
  }

  switch(ChangeDisplaySettings(&devmode, CDS_TEST)) {
  case DISP_CHANGE_SUCCESSFUL:
  case DISP_CHANGE_NOTUPDATED:
    // Successful change of resolution. (note: DISP_CHANGE_NOTUPDATED may
    // occur under NT/2000 if the user is non-Administrator, but resolution
    // is changed anyway)
    ChangeDisplaySettings (&devmode, CDS_FULLSCREEN);
    return TRUE;
  case DISP_CHANGE_RESTART:
    MessageBox(0, "Cannot run in 16-bit without restart", APP_NAME, MB_ICONERROR);
    return 0;
  case DISP_CHANGE_BADMODE:
    MessageBox(0, "640x480 fullscreen 16-bit not supported", APP_NAME, MB_ICONERROR);
    return 0;
  case DISP_CHANGE_FAILED:
    MessageBox(0, "Hardware failed to change to 640x480x16bit", APP_NAME, MB_ICONERROR);
    return 0;
  default:
    MessageBox(0, "Unknown error when changing to 640x480x16bit", APP_NAME, MB_ICONERROR);
    return 0;
  }
#endif
}

// Return to the original display mode.
void restore_display_mode(void)
{
#if USE_DIRECTDRAW
  if (g_DirectDraw)  g_DirectDraw->Release();
  g_DirectDraw = NULL;
#else
  ChangeDisplaySettings(NULL, 0);
#endif
}

// Initialise an OpenGL rendering context for the window and bind to it.
HGLRC init_opengl(HWND hWnd)
{
  int  pixelFormat;
  static PIXELFORMATDESCRIPTOR pfd = {
        sizeof(PIXELFORMATDESCRIPTOR),  // size of this pfd
        1,                              // version number
        PFD_DRAW_TO_WINDOW              // support window
        |  PFD_SUPPORT_OPENGL           // support OpenGL
        |  PFD_DOUBLEBUFFER ,           // double buffered
        PFD_TYPE_RGBA,                  // RGBA type
        16,                             // 16-bit color depth
        0, 0, 0, 0, 0, 0,               // color bits ignored
        0,                              // no alpha buffer
        0,                              // shift bit ignored
        0,                              // no accumulation buffer
        0, 0, 0, 0,                     // accum bits ignored
        16,                             // 16-bit z-buffer      
        0,                              // no stencil buffer
        0,                              // no auxiliary buffer
        PFD_MAIN_PLANE,                 // main layer
        0,                              // reserved
        0, 0, 0                         // layer masks ignored
    };

  app_hdc = GetDC(hWnd);

  pixelFormat = ChoosePixelFormat(app_hdc, &pfd);
  if (!pixelFormat) {
    MessageBox(0, "Couldn't find appropriate pixelformat", APP_NAME, MB_ICONERROR);
    return 0;
  }

  if (!SetPixelFormat(app_hdc, pixelFormat, &pfd)) {
    MessageBox(0, "Couldn't bind pixelformat to window", APP_NAME, MB_ICONERROR);
    return 0;
  }

  app_hglrc = wglCreateContext(app_hdc);
  if (!app_hglrc) {
    MessageBox(0, "wglCreateContext failed", APP_NAME, MB_ICONERROR);
    return 0;
  }

  if (!wglMakeCurrent(app_hdc, app_hglrc)) {
    wglDeleteContext(app_hglrc);
    app_hglrc = 0;

    MessageBox(0, "Couldn't make GL rendering context current", APP_NAME, MB_ICONERROR);
    return 0;
  }

  return app_hglrc;
}

// Shut down OpenGL.
void shutdown_opengl(void)
{
  if (app_hdc)   ReleaseDC(app_hwnd, app_hdc);
  if (app_hglrc) wglDeleteContext(app_hglrc);

  app_hdc = 0;
  app_hglrc = 0;
}

// Start up DirectInput.
int init_directinput(HINSTANCE hInstance, HWND hWnd)
{
  HRESULT hr;

  // Get the DirectInput interface.
  hr = DirectInputCreate(hInstance, DIRECTINPUT_VERSION, &g_DirectInput, NULL);
  if (hr == DIERR_OLDDIRECTINPUTVERSION) {
    MessageBox(0, "Need a newer version of DirectInput (requires DirectX 3).", APP_NAME, MB_ICONERROR);
    return 0;
  } else if (hr != DI_OK) {
    MessageBox(0, "Couldn't bind to DirectInput", APP_NAME, MB_ICONERROR);
    return 0;
  }

  // Create the keyboard device.
  hr = g_DirectInput->CreateDevice(GUID_SysKeyboard, &g_DIKeyboard, NULL);
  if (FAILED(hr)) {
    MessageBox(0, "Can't create SysKeyboard DirectInput device", APP_NAME, MB_ICONERROR);
    return 0;
  }

  hr = g_DIKeyboard->SetDataFormat(&c_dfDIKeyboard);
  if (FAILED(hr)) {
    MessageBox(0, "Can't set Keyboard input type on g_DIKeyboard", APP_NAME, MB_ICONERROR);
    return 0;
  }

  // Set the keyboard's cooperative level so that we have non-exclusive access
  // (can't get exclusive access on keyboards) in the foreground.
  hr = g_DIKeyboard->SetCooperativeLevel(hWnd, DISCL_FOREGROUND | DISCL_NONEXCLUSIVE);
  if (FAILED(hr)) {
    MessageBox(0, "Can't set Keyboard's CooperativeLevel", APP_NAME, MB_ICONERROR);
    return 0;
  }

  // Acquire the keyboard.
  hr = g_DIKeyboard->Acquire();
  if (FAILED(hr)) {
    MessageBox(0, "Can't Acquire DIKeyboard", APP_NAME, MB_ICONERROR);
    return 0;
  }


  // Create the mouse device.
  hr = g_DirectInput->CreateDevice(GUID_SysMouse, &g_DIMouse, NULL);
  if (FAILED(hr)) {
    MessageBox(0, "Can't create SysMouse DirectInput device", APP_NAME, MB_ICONERROR);
    return 0;
  }

  hr = g_DIMouse->SetDataFormat(&c_dfDIMouse);
  if (FAILED(hr)) {
    MessageBox(0, "Can't set Keyboard input type on g_DIMouse", APP_NAME, MB_ICONERROR);
    return 0;
  }

  // Set the mouse's cooperative level so that we have exclusive access
  // when in the foreground. This has the side-effect of causing the
  // mouse pointer to disappear.
  hr = g_DIMouse->SetCooperativeLevel(hWnd, DISCL_EXCLUSIVE | DISCL_FOREGROUND);
  if (FAILED(hr)) {
    MessageBox(0, "Can't set Mouse's CooperativeLevel", APP_NAME, MB_ICONERROR);
    return 0;
  }

  // Set the mouse to absolute mode.
  DIPROPDWORD diProp;

  memset((void *) &diProp, 0, sizeof(diProp));
  diProp.diph.dwSize = sizeof(diProp);
  diProp.diph.dwHeaderSize = sizeof(DIPROPHEADER);
  diProp.diph.dwHow = DIPH_DEVICE;
  diProp.dwData = DIPROPAXISMODE_ABS;
  hr = g_DIMouse->SetProperty(DIPROP_AXISMODE, &diProp.diph);
  if (FAILED(hr)) {
    MessageBox(0, "Can't set Mouse to Absolute axis mode", APP_NAME, MB_ICONERROR);
    return 0;
  }

  // Acquire the mouse.
  hr = g_DIMouse->Acquire();
  if (FAILED(hr)) {
    MessageBox(0, "Can't Acquire DIMouse", APP_NAME, MB_ICONERROR);
    return 0;
  }

  return TRUE;
}

// Shut down DirectInput.
void shutdown_directinput(void)
{
  if (g_DIKeyboard)  { g_DIKeyboard->Unacquire(); g_DIKeyboard->Release(); }
  if (g_DirectInput) g_DirectInput->Release();

  g_DIKeyboard = NULL;
  g_DirectInput = NULL;
}

// Poll the keyboard - places keyboard's current state into 256-byte kbd_buffer.
void poll_keyboard(unsigned char *kbd_buffer)
{
  HRESULT hr;
  
  hr = g_DIKeyboard->GetDeviceState(256, (LPVOID) kbd_buffer);
  if ((hr == DIERR_INPUTLOST) || (hr == DIERR_NOTACQUIRED)) {
    // Try to re-acquire the keyboard.
    hr = g_DIKeyboard->Acquire();
    if (hr == DI_OK) {
      // Okay, we have got it again. Poll the keyboard once more.
      hr = g_DIKeyboard->GetDeviceState(256, (LPVOID) kbd_buffer);
    }
  }

  if (FAILED(hr)) {
    // Couldn't sample the keyboard; don't change the state of the buffer.
    return;
  }
}

// Poll the mouse - places mouse's current state into coords[2] and buttons[2].
void poll_mouse(long *coords, unsigned char *buttons)
{
  HRESULT hr;
  DIMOUSESTATE mouseState;
  
  hr = g_DIMouse->GetDeviceState(sizeof(mouseState), (LPVOID) &mouseState);
  if ((hr == DIERR_INPUTLOST) || (hr == DIERR_NOTACQUIRED)) {
    // Try to re-acquire the mouse.
    hr = g_DIMouse->Acquire();
    if (hr == DI_OK) {
      // Okay, we have got it again. Poll the mouse once more.
      hr = g_DIMouse->GetDeviceState(sizeof(mouseState), (LPVOID) &mouseState);
    }
  }

  if (SUCCEEDED(hr)) {
    // Sampled the mouse successfully.
    coords[0] = mouseState.lX;
    coords[1] = mouseState.lY;
    buttons[0] = mouseState.rgbButtons[0];
    buttons[1] = mouseState.rgbButtons[1];
  }
}

// Open the debugging logfile.
int init_debug_file(void)
{
  debug_file = fopen("debuglog.txt", "w");
  if (!debug_file) {
    MessageBox(0, "Couldn't open debugging logfile", APP_NAME, MB_ICONERROR);
    return 0;
  }

  return TRUE;
}

void close_debug_file(void)
{
  if (debug_file) {
    fclose(debug_file);
    debug_file = NULL;
  }
}


// Entry point for the application.
int APIENTRY WinMain(HINSTANCE hInstance,
                     HINSTANCE hPrevInstance,
                     LPSTR     lpCmdLine,
                     int       nCmdShow)
{
  MSG msg;

  app_hinstance = hInstance;

  if (!init_debug_file())
    return 0;

  // First: call the application-defined routines
  // that do their thing before the main GL window comes up.
  user_init();

  /*
   * Initialisation
   */

  if (!init_window_class(hInstance))
    return 0;

  if (!init_main_window(hInstance))
    return 0;

  if (view_fullscreen)
    if (!init_display_mode())
      return 0;

  ShowCursor(FALSE);
  SetForegroundWindow(app_hwnd);
  SetWindowPos(app_hwnd, HWND_TOPMOST, 0, 0, viewresolution_x, viewresolution_y, 0);
 
  if (!init_opengl(app_hwnd))
    DestroyWindow(app_hwnd);

  if (!init_directinput(hInstance, app_hwnd))
    return 0;

  requires_shutdown = TRUE;

  // Initialise the user application.
  user_setup();

  // We require accurate timing info (1ms)
  timeBeginPeriod(1);

  while (1) {
    while (PeekMessage(&msg, 0, 0, 0, PM_REMOVE)) {
      TranslateMessage(&msg);
			DispatchMessage(&msg);
    }
    if (exit_application)
      break;

    if (user_idle())
      InvalidateRect(app_hwnd, NULL, FALSE);
  }

  user_destroy();
  system_cleanup();
  close_debug_file();
	return 0;
}

// Cleanup all our connections to the system.
void system_cleanup(void)
{
  // Show the cursor again.
  ShowCursor(TRUE);

  // No longer require the accurate timing info.
  timeEndPeriod(1);

  shutdown_directinput();
  shutdown_opengl();
  if (view_fullscreen)
    restore_display_mode();
}

// Initiate a premature shutdown - the user code has caused a fatal error.
void premature_shutdown(void)
{
  if (requires_shutdown)
    system_cleanup();

  close_debug_file();
  exit(-1);
}

// Main window message handler.
LONG WINAPI MainWndProc(HWND    hWnd, UINT    uMsg,
                        WPARAM  wParam, LPARAM  lParam)
{
  LONG returncode = 1;

  switch (uMsg) {
    case WM_CREATE:
      break;

    case WM_SIZE:
      // We are running fullscreen and hopefully shouldn't get any of these requests.
      break;

    case WM_CLOSE:
      DestroyWindow(hWnd);
      break;

    case WM_DESTROY:
      PostQuitMessage(0);
      exit_application = TRUE;
      break;

    case WM_QUIT:
      // On our next loop through WinMain, exit (posting the quit message)
      exit_application = TRUE;
      break;

    case WM_PAINT: {
      PAINTSTRUCT ps;

      BeginPaint(hWnd, &ps);

      // Call the user's drawing routine.
      user_redraw();

      // Slap it onto the screen.
      glFlush();
      SwapBuffers(app_hdc);

      EndPaint(hWnd, &ps);

      last_frame_displayed = timeGetTime();
      break;
    }

    default:
      returncode = DefWindowProc(hWnd, uMsg, wParam, lParam);
      break;
  }

  return returncode;
}

// End the program. Called by the user code.
void end_program(void)
{
  if (requires_shutdown)
    SendMessage(app_hwnd, WM_CLOSE, 0, 0);
  else
    exit(0);
}

// File selection routine - TRUE on success
int choose_file(int filenamebufsize, char *filenamebuf, char *msg)
{
  OPENFILENAME ofn;

  filenamebuf[0] = '\0';

  memset((void *) &ofn, 0, sizeof(OPENFILENAME));
  ofn.lStructSize = sizeof(ofn);
  ofn.Flags = OFN_FILEMUSTEXIST | OFN_HIDEREADONLY | OFN_LONGNAMES | OFN_NOCHANGEDIR | OFN_PATHMUSTEXIST;
  ofn.hwndOwner = app_hwnd;
  ofn.lpstrFile = filenamebuf;
  ofn.nMaxFile = filenamebufsize;
  ofn.lpstrTitle = (LPCTSTR) msg;

  return GetOpenFileName(&ofn);
}

// Directory traversal routines
int iterate_directory(void **directory, char *dirpath, char *return_filename, int *return_isdir)
{
  WIN32_FIND_DATA findData;
  dirsearchstruct *searchstruct = (dirsearchstruct *) *directory;
  BOOL found;

  if (searchstruct == NULL) {
    char searchnamebuf[256];
    int len = strlen(dirpath);

    // Add a trailing \ to the directory if there isn't one.
    if (dirpath[len - 1] != '\\') {
      dirpath[len++] = '\\';
      dirpath[len] = '\0';
    }
    sprintf(searchnamebuf, "%s*.*", dirpath);

    searchstruct = (dirsearchstruct *) malloc(sizeof(dirsearchstruct));
    if (!searchstruct) mem_fail(sizeof(dirsearchstruct), "directory-search structure");
    *directory = (void *) searchstruct;
    searchstruct->hFindFile = FindFirstFile(searchnamebuf, &findData);
    if (searchstruct->hFindFile == INVALID_HANDLE_VALUE)
      return FALSE;

    found = TRUE;
  } else {
    found = FindNextFile(searchstruct->hFindFile, &findData);
  }

  if (!found)
    return FALSE;

  *return_isdir = (findData.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY);
  sprintf(return_filename, "%s%s", dirpath, findData.cFileName);

  return TRUE;
}

void end_directory_traverse(void **directory)
{
  dirsearchstruct *searchstruct = (dirsearchstruct *) *directory;

  if (searchstruct->hFindFile != INVALID_HANDLE_VALUE)
    FindClose(searchstruct->hFindFile);

  free(searchstruct);
}

void mem_fail(size_t size_allocated, char *memtype)
{
  if (size_allocated == 0) {
    fprintf(debug_file, "WARNING: tried to allocate 0 bytes for '%s'\n", memtype);
    return;
  }

  fprintf(debug_file, "OUT OF MEMORY: allocation of %d bytes for '%s' failed!\n",
          size_allocated, memtype);
  premature_shutdown();
}