// CoreTag.cpp : Routines for class CoreTag and supporting code.
//

#include "stdafx.h"


// Create a new tag.
CoreTag::CoreTag(MythPlugin *my_plugin, tagheader *header_struct)
  : Tag(my_plugin, header_struct)
{
  sprite_collection = NULL;
}

// Read the core data from disk.
void CoreTag::read_data(void)
{
  int i;

  // Seek the tag-file to the start of the tag's data.
  fseek(plugin->tagfile, header->offset, SEEK_SET);
//  fprintf(debug_file, "read core '%s' from %s @ %d\n",
//          tagid_to_string(header->our_id), plugin->header.name, header->offset);

  // Read the header for the core data.
  data.x256_id = myth_read_tag(plugin->tagfile);
//  fprintf(debug_file, "  .256 tag '%s'\n", tagid_to_string(data.x256_id));

  data.num_entries = myth_read_short(plugin->tagfile);
//  fprintf(debug_file, "  %d entries\n", data.num_entries);

  data.tintval = myth_read_short(plugin->tagfile);
//  fprintf(debug_file, "  tintval %d\n", data.tintval);

  // Read the core entries.
  for (i = 0; i < 5; i++) {
//    fprintf(debug_file, "    entry %d:\n", i);

    myth_read_bytes(plugin->tagfile, 8, data.entry[i].colors[0]);
//    fprintf(debug_file, "      color 0:");
//    for (j = 0; j < 8; j++) {
//      fprintf(debug_file, " %02X", data.entry[i].colors[0][j]);
//    }
//    fprintf(debug_file, "\n");

    myth_read_bytes(plugin->tagfile, 8, data.entry[i].colors[1]);
//    fprintf(debug_file, "      color 1:");
//    for (j = 0; j < 8; j++) {
//      fprintf(debug_file, " %02X", data.entry[i].colors[1][j]);
//    }
//    fprintf(debug_file, "\n");

    myth_read_bytes(plugin->tagfile, 8, data.entry[i].colors[2]);
//    fprintf(debug_file, "      color 2:");
//    for (j = 0; j < 8; j++) {
//      fprintf(debug_file, " %02X", data.entry[i].colors[2][j]);
//    }
//    fprintf(debug_file, "\n");

    myth_read_longs(plugin->tagfile, 10, data.entry[i].unknown);
//    for (j = 0; j < 10; j++)
//      fprintf(debug_file, "      unknown[%d]: 0x%08X\n", j, data.entry[i].unknown[j]);
  }

  myth_read_bytes(plugin->tagfile, 8, data.tintcol);
//  fprintf(debug_file, "  tint color: ");
//  for (j = 0; j < 8; j++) {
//    fprintf(debug_file, " %02X", data.tintcol[j]);
//  }
//  fprintf(debug_file, "\n");

  myth_read_longs(plugin->tagfile, 10, data.unknown2);
//  for (j = 0; j < 10; j++)
//    fprintf(debug_file, "  unknown2[%d]: 0x%08X\n", j, data.unknown2[j]);

  loaded_data = TRUE;
}

// Read the bulk mesh data (the big arrays) from disk.
void CoreTag::load_bulk_data(void)
{
  // Do nothing.
  bulkdata_size = 0;
  loaded_bulk_data = TRUE;
}

// Free the big mesh arrays that we don't
// need any more (but keep the rest of the tag around).
void CoreTag::free_bulk_data(void)
{
  // Do nothing.
}

// Set up the collection-reference's texture in the bintree.
void CoreTag::setup_bintree_textures(Bintree *bintree, int texture_index,
                                     int bitmap_id, short scale_factor)
{
  TagHashStructure *hash;
  int i;

  // Read our data.
//  fprintf(debug_file, "core '%s' loaded_data %d\n", tagid_to_string(header->our_id), loaded_data);
  if (!loaded_data)
    this->read_data();

  // Get the .256 hash structure.
  hash = myth_tags->hash_structure(string_to_tagid(".256"));
  assert(hash != NULL);

  // Find our sprite collection.
  sprite_collection = (X256Tag *) hash->retrieve_tag(data.x256_id);
  if (sprite_collection == NULL) {
    fprintf(debug_file, "ERROR: collection reference '%s' couldn't find sprite collection '%s' (crosscheck: '%s')\n",
            tagid_to_string(header->our_id), tagid_to_string(data.x256_id), tagid_to_string((tag_id) header->userdata));
    premature_shutdown();
  }

  // Set up the bintree's texturing from the sprite, for every collection reference entry.
  for (i = 0; i < data.num_entries; i++)
    sprite_collection->bintree_sprite_texture(bintree, texture_index + i,
                                              bitmap_id, scale_factor, 
                                              data.tintval, data.tintcol,
                                              &(data.entry[i]));
}

// Dump our sprite data.
void CoreTag::dump_sprite_data(void)
{
  if ((sprite_collection != NULL) && (sprite_collection->loaded_bulk_data))
    sprite_collection->free_bulk_data();
}
