// CoreTag.h : Definitions for class CoreTag and supporting code.
//

// The 64-byte entry in the collection reference that sets up
// colors to change to.
typedef struct coreentry_ {
  unsigned char colors[3][8];// Three colors here, each eight bytes.
                             // Each color is RRGGBBXX where X = ?
  long unknown[10];
} coreentry;

// The 384-byte collection reference data.
typedef struct coredata_ {
  tag_id x256_id;          // sprite collection ref (of type '.256')
  short num_entries;       // number of entry structures filled
  short tintval;           // 0-256 is 0 - 100%
  coreentry entry[5];      // 64-byte entries
  unsigned char tintcol[8];// same structure as coreentry colors
  long unknown2[10];       // filler
} coredata;

// A core - a collection reference.
class CoreTag : public Tag {
public:
  coredata data;
  class X256Tag *sprite_collection;

  // Read the core's static data from disk.
  virtual void read_data(void);

  // These virtual functions do nothing.
  virtual void load_bulk_data(void);
  virtual void free_bulk_data(void);

  // Set up the core's texture in the bintree.
  void setup_bintree_textures(Bintree *bintree, int texture_index,
                              int bitmap_id, short scale_factor);

  // Dump our sprite data.
  void dump_sprite_data(void);

  // Create a new tag.
  CoreTag(class MythPlugin *plugin, tagheader *header);

  // Delete the tag.
  ~CoreTag();
};
