// Dialog.cpp : code for the Map and Hello dialogs
//

#include "stdafx.h"

BOOL CALLBACK MapDlgProc(HWND hwnd, UINT Message, WPARAM wParam, LPARAM lParam)
{
  int i, selected;

  switch(Message) {
  case WM_INITDIALOG:
    // Add the maps' names to the drop-down combobox. Select the first one by default.
    for (i = 0; i < mapsel_num_entrypoints; i++)
      SendDlgItemMessage(hwnd, IDC_MAPCOMBO, CB_ADDSTRING, (WPARAM) 0, (LPARAM) mapsel_entrypoint[i]->meshname);
    SendDlgItemMessage(hwnd, IDC_MAPCOMBO, CB_SETCURSEL, 0, 0);

    return TRUE;

  case WM_COMMAND:
    switch(LOWORD(wParam)) {
    case IDOK:
      selected = SendDlgItemMessage(hwnd, IDC_MAPCOMBO, CB_GETCURSEL, 0, 0);
      if (selected == CB_ERR) {
        MessageBox(hwnd, "You must select a map to view.", "Select a Mesh",
                   MB_OK | MB_ICONINFORMATION);
      } else {
        mapsel_choice = selected;
        EndDialog(hwnd, IDOK);
      }
      return TRUE;
    case IDCANCEL:
      mapsel_choice = -1;
      EndDialog(hwnd, IDCANCEL);
      return TRUE;
    case IDC_MAPCOMBO:
      switch(HIWORD(wParam)) {
      case CBN_DBLCLK:
        // The user has double-clicked an item. Close the dialog box.
        mapsel_choice = SendDlgItemMessage(hwnd, IDC_MAPCOMBO, CB_GETCURSEL, 0, 0);
        EndDialog(hwnd, IDOK);
        return TRUE;
      }
      break;
    }
    break;
  }

  return FALSE;
}

BOOL CALLBACK HelloDlgProc(HWND hwnd, UINT Message, WPARAM wParam, LPARAM lParam)
{
  switch(Message) {
  case WM_INITDIALOG:
    // Set up the dialog box.
    SetDlgItemText(hwnd, IDC_README, "PluginViewer - an OpenGL toy\r\n"
                                     "Chris Butcher, 1999\r\n"
                                     "\r\n"
                                     "  IMPORTANT: The next dialog box will ask you\r\n"
                                     "to navigate to Myth II.exe. Don't panic, this won't\r\n"
                                     "trash your game in any way - it just needs to know\r\n"
                                     "where your Myth II directory is so that it can find\r\n"
                                     "your plugins. Please select Myth II.exe rather than\r\n"
                                     "trying to pick a specific plugin to view - that part\r\n"
                                     "comes later.\r\n"
                                     "\r\n"
                                     "  Briefly, this app reads Myth II: Soulblighter maps\r\n"
                                     "and lets you fly around them in real-time. Use the\r\n"
                                     "mouse to move around, A to go forwards, Z to go\r\n"
                                     "backwards, Q and E to roll the camera. See readme.txt\r\n"
                                     "for more detailed instructions.\r\n"
                                     "\r\n"
                                     "  Requires an OpenGL-compatible 3D accelerator.\r\n"
                                     "\r\n"
                                     "  This program is email-ware; if you like it, send\r\n"
                                     "me a note at cbutcher@cs.otago.ac.nz. Source code\r\n"
                                     "is available upon request.\r\n");

    return TRUE;

  case WM_COMMAND:
    switch(LOWORD(wParam)) {
    case IDOK:
      EndDialog(hwnd, IDOK);
      return TRUE;
      break;
    }
  }

  return FALSE;
}

// Raise the map selection dialog box.
void map_selection_dialog(void)
{
  int mapdialog_return = DialogBox(app_hinstance, MAKEINTRESOURCE(IDD_MAP), 0, MapDlgProc);

  if (mapdialog_return == IDCANCEL) {
    mapsel_choice = -1;
  } else if (mapdialog_return == -1) {
    fprintf(debug_file, "ERROR: Map Selection dialog returned an error!\n");
    premature_shutdown();
  }
}

// Show the hello box.
void show_hello_box(void)
{
  DialogBox(app_hinstance, MAKEINTRESOURCE(IDD_HELLO), 0, HelloDlgProc);
}