// Init.h : Application initialisation and setup.
//

/*
 * Global settings... must be initialised by user app in user_init.
 */

extern unsigned int viewresolution_x, viewresolution_y;
extern BOOL view_fullscreen;

#define APP_NAME     "TerrainApp"
#define MAX_ENUMS    50

/*
 * Global variables
 */

extern BOOL      exit_application;
extern HINSTANCE app_hinstance;
extern HWND      app_hwnd;
extern HDC       app_hdc;
extern HGLRC     app_hglrc;
extern FILE     *debug_file;

extern LPDIRECTINPUT       g_DirectInput;
extern LPDIRECTINPUTDEVICE g_DIKeyboard, g_DIMouse;

/*
 * Function prototypes
 */

// Register the window class.
int init_window_class(HINSTANCE hInstance);

// Create, show and focus the application's main window.
int init_main_window(HINSTANCE hInstance);

// Change display modes.
int init_display_mode(void);
void restore_display_mode(void);

// Open and close the debugging logfile
int init_debug_file(void);
void close_debug_file(void);

// Set up and shut down an OpenGL rendering context.
HGLRC init_opengl(HWND hWnd);
void shutdown_opengl(void);

// Set up and shut down the DirectInput connection.
int init_directinput(HINSTANCE hInstance, HWND hWnd);
void shutdown_directinput(void);

// Entry point.
int APIENTRY WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance,
                     LPSTR     lpCmdLine, int       nCmdShow);

// Main window message handler.
LONG WINAPI MainWndProc(HWND    hWnd, UINT    uMsg,
                        WPARAM  wParam, LPARAM  lParam);

// End the program. Called by the user code.
void end_program(void);

// Cleanup all our connections to the system.
void system_cleanup(void);

// Initiate a premature shutdown - the user code has caused a fatal error.
void premature_shutdown(void);

// Poll the keyboard - places keyboard's current state into 256-byte kbd_buffer.
void poll_keyboard(unsigned char *kbd_buffer);

// Poll the mouse - places mouse's current state into coords[2] and buttons[2].
void poll_mouse(long *coords, unsigned char *buttons);

// File selection routine - TRUE on success
int choose_file(int filenamebufsize, char *filenamebuf, char *msg);

// Directory traversal routines
typedef struct dirsearchstruct_ {
  HANDLE hFindFile;
} dirsearchstruct;

int iterate_directory(void **directory, char *dirpath, char *return_filename, int *return_isdir);
void end_directory_traverse(void *directory);

// Memory failure.
void mem_fail(size_t size_allocated, char *memtype);