// MeshTag.cpp : Routines for class MeshTag and supporting code.
//

#include "stdafx.h"


// Create a new tag.
MeshTag::MeshTag(MythPlugin *my_plugin, tagheader *header_struct)
  : Tag(my_plugin, header_struct)
{
}

// Read the mesh data from disk.
void MeshTag::read_data(void)
{
  // Seek the tag-file to the start of the tag's data.
  fseek(plugin->tagfile, header->offset, SEEK_SET);

  // Read the 1024-byte mesh header.
  mesh_header = (meshheader *) malloc(sizeof(meshheader));
  if (!mesh_header) mem_fail(sizeof(meshheader),
                             "MeshTag header");

  mesh_header->colormap = myth_read_tag(plugin->tagfile);
  mesh_header->watertype = myth_read_tag(plugin->tagfile);

  mesh_header->x_sections = myth_read_short(plugin->tagfile);
  mesh_header->y_sections = myth_read_short(plugin->tagfile);

  mesh_header->unknown1 = myth_read_long(plugin->tagfile);
  mesh_header->height_length = myth_read_long(plugin->tagfile);
  mesh_header->unknown2 = myth_read_long(plugin->tagfile);
  mesh_header->bulk_offset = myth_read_long(plugin->tagfile);
  mesh_header->bulk_datasize = myth_read_long(plugin->tagfile);

  mesh_header->unknown3 = myth_read_long(plugin->tagfile);
  mesh_header->num_itemtypes = myth_read_long(plugin->tagfile);
  mesh_header->itemtype_offset = myth_read_long(plugin->tagfile);
  mesh_header->itemtype_length = myth_read_long(plugin->tagfile);

  mesh_header->unknown4 = myth_read_long(plugin->tagfile);
  mesh_header->num_items = myth_read_long(plugin->tagfile);
  mesh_header->item_offset = myth_read_long(plugin->tagfile);
  mesh_header->item_length = myth_read_long(plugin->tagfile);
  
  mesh_header->unknown5 = myth_read_long(plugin->tagfile);
  mesh_header->meshlighting = myth_read_tag(plugin->tagfile);
  mesh_header->watertype2 = myth_read_tag(plugin->tagfile);

  mesh_header->mesh_flags = myth_read_short(plugin->tagfile);
  mesh_header->netgame_flags = myth_read_short(plugin->tagfile);

  mesh_header->particle = myth_read_tag(plugin->tagfile);
  mesh_header->num_teams = myth_read_long(plugin->tagfile);

  mesh_header->unknown6 = myth_read_long(plugin->tagfile);
  mesh_header->unknown7 = myth_read_long(plugin->tagfile);
  mesh_header->unknown8 = myth_read_long(plugin->tagfile);
  mesh_header->unknown9 = myth_read_long(plugin->tagfile);
  mesh_header->unknown10 = myth_read_long(plugin->tagfile);
  mesh_header->unknown11 = myth_read_long(plugin->tagfile);
  mesh_header->unknown12 = myth_read_long(plugin->tagfile);
  mesh_header->unknown13 = myth_read_short(plugin->tagfile);
  mesh_header->unknown14 = myth_read_short(plugin->tagfile);
  mesh_header->unknown15 = myth_read_short(plugin->tagfile);
  mesh_header->unknown16 = myth_read_short(plugin->tagfile);

  mesh_header->ambientsound = myth_read_tag(plugin->tagfile);

  mesh_header->num_actions = myth_read_long(plugin->tagfile);
  mesh_header->action_offset = myth_read_long(plugin->tagfile);
  mesh_header->action_length = myth_read_long(plugin->tagfile);

  mesh_header->mapname = myth_read_tag(plugin->tagfile);
  mesh_header->postgame = myth_read_tag(plugin->tagfile);
  mesh_header->pregame = myth_read_tag(plugin->tagfile);
  mesh_header->overhead = myth_read_tag(plugin->tagfile);
  mesh_header->nextlevel = myth_read_tag(plugin->tagfile);

  mesh_header->unknown17 = myth_read_long(plugin->tagfile);
  mesh_header->objectives = myth_read_tag(plugin->tagfile);
  mesh_header->unknown18 = myth_read_long(plugin->tagfile);
  mesh_header->unknown19 = myth_read_long(plugin->tagfile);
  mesh_header->narration = myth_read_tag(plugin->tagfile);
  mesh_header->unknown20 = myth_read_long(plugin->tagfile);
  mesh_header->unknown21 = myth_read_long(plugin->tagfile);
  mesh_header->unknown22 = myth_read_long(plugin->tagfile);

  mesh_header->array1_offset = myth_read_long(plugin->tagfile);
  mesh_header->array1_length = myth_read_long(plugin->tagfile);

  mesh_header->unknown23 = myth_read_long(plugin->tagfile);
  mesh_header->array2_offset = myth_read_long(plugin->tagfile);
  mesh_header->array2_length = myth_read_long(plugin->tagfile);

  mesh_header->unknown24 = myth_read_long(plugin->tagfile);
  mesh_header->unknown25 = myth_read_long(plugin->tagfile);
  mesh_header->unknown26 = myth_read_long(plugin->tagfile);
  mesh_header->unknown27 = myth_read_long(plugin->tagfile);

  mesh_header->wind = myth_read_tag(plugin->tagfile);

  mesh_header->unknown28 = myth_read_long(plugin->tagfile);
  mesh_header->unknown29 = myth_read_long(plugin->tagfile);
  mesh_header->unknown30 = myth_read_long(plugin->tagfile);
  mesh_header->unknown31 = myth_read_long(plugin->tagfile);
  mesh_header->unknown32 = myth_read_long(plugin->tagfile);

  mesh_header->subtitle = myth_read_tag(plugin->tagfile);
  mesh_header->narrationsound = myth_read_tag(plugin->tagfile);

  myth_read_longs(plugin->tagfile, 191, mesh_header->unknown33);
  
  loaded_data = TRUE;
}

// Read the bulk mesh data (the big arrays) from disk.
void MeshTag::load_bulk_data(void)
{
  int i, j, grid_x, grid_y, gridsize;
  gridval *gridptr;
  itemtype *itptr;
  itempos *iptr;

  if (!loaded_bulk_data) {
    bulkdata_size = 0;
    loaded_bulk_data = TRUE;

    // Allocate memory for the grid data. There are 32 samples per
    // mesh section in X and Y.
    grid_x = mesh_header->x_sections * 32;
    grid_y = mesh_header->y_sections * 32;
    gridsize = grid_x * grid_y;
    grid = (gridval *) malloc(gridsize * sizeof(gridval));
    if (!grid) mem_fail(gridsize * sizeof(gridval),
                        "MeshTag height array");

    // Seek to the correct location in the tagfile.
    fseek(plugin->tagfile, header->offset + mesh_header->bulk_offset, SEEK_SET);

    max_height = min_height = 0;

    for (i = 0, gridptr = grid; i < gridsize; i++, gridptr++) {
      gridptr->height = myth_read_short(plugin->tagfile);
      gridptr->grad_x = myth_read_byte(plugin->tagfile);
      gridptr->grad_y = myth_read_byte(plugin->tagfile);
      gridptr->water_flags = myth_read_byte(plugin->tagfile);
      gridptr->passability = myth_read_byte(plugin->tagfile);
      for (j = 0; j < 3; j++)
        gridptr->unused[j] = myth_read_short(plugin->tagfile);

      if (gridptr->height < min_height)
        min_height = gridptr->height;
      else if (gridptr->height > max_height)
        max_height = gridptr->height;
    }

    bulkdata_size += gridsize * sizeof(gridval);

    num_scenery_types = 0;
    if (mesh_header->num_itemtypes > 0) {
      // Allocate the item-type array.
      itemtypearray = (itemtype *) malloc(mesh_header->num_itemtypes * sizeof(itemtype));
      if (!itemtypearray) mem_fail(mesh_header->num_itemtypes * sizeof(itemtype),
                                   "MeshTag itemtype array");

      // Allocate a reverse lookup table to convert itemtype_index (which is into the
      // number of scenery objects) into a _real_ index into the itemtype buffer.
      // Why Bungie did this I have no idea.
      scenery_lookup = (int *) malloc(mesh_header->num_itemtypes * sizeof(int));
      if (!scenery_lookup) mem_fail(mesh_header->num_itemtypes * sizeof(int),
                                    "MeshTag scenery-lookup array");
      memset(scenery_lookup, 0xFF, mesh_header->num_itemtypes * sizeof(int));
      
      // Seek to the correct location in the tagfile.
      fseek(plugin->tagfile, header->offset + mesh_header->bulk_offset + mesh_header->itemtype_offset, SEEK_SET);

      for (i = 0, itptr = itemtypearray; i < mesh_header->num_itemtypes; i++, itptr++) {
        itptr->type = myth_read_short(plugin->tagfile);
        itptr->flags = myth_read_short(plugin->tagfile);
        itptr->item_tag = myth_read_tag(plugin->tagfile);
        itptr->team = myth_read_short(plugin->tagfile);
        
        myth_read_shorts(plugin->tagfile, 5, itptr->unknown);

        // Read a long and discard it.
        myth_read_long(plugin->tagfile);
        itptr->scenerytag = NULL;
        
        myth_read_short(plugin->tagfile);
        itptr->bintree_scenery_index = -1;
        
        itptr->item_count = myth_read_short(plugin->tagfile);        
        itptr->item_max = myth_read_short(plugin->tagfile);
        itptr->itemtype_index = myth_read_short(plugin->tagfile);

        if (itptr->type == ITEMTYPE_SCENERY)
          scenery_lookup[num_scenery_types++] = i;
      }
      
      bulkdata_size += mesh_header->num_itemtypes * (sizeof(itemtype) + sizeof(int));
    } else {
      itemtypearray = NULL;
    }

    num_scenery_items = 0;
    if (mesh_header->num_items > 0) {
      // Allocate the item-position array.
      itemarray = (itempos *) malloc(mesh_header->num_items * sizeof(itempos));
      if (!itemarray) mem_fail(mesh_header->num_items * sizeof(itempos),
                               "MeshTag item-position array");
      
      // Seek to the correct location in the tagfile.
      fseek(plugin->tagfile, header->offset + mesh_header->bulk_offset + mesh_header->item_offset, SEEK_SET);

      for (i = 0, iptr = itemarray; i < mesh_header->num_items; i++, iptr++) {
        iptr->unknown1 = myth_read_short(plugin->tagfile);
        iptr->flags = myth_read_short(plugin->tagfile);

        iptr->type = myth_read_short(plugin->tagfile);
        if (iptr->type == ITEMTYPE_SCENERY)
          num_scenery_items++;

        iptr->itemtype_index = myth_read_short(plugin->tagfile);
        iptr->item_id = myth_read_short(plugin->tagfile);
        iptr->unknown2 = myth_read_short(plugin->tagfile);

        myth_read_longs(plugin->tagfile, 3, iptr->item_pos);

        myth_read_shorts(plugin->tagfile, 4, iptr->unknown3);
        iptr->orientation = myth_read_short(plugin->tagfile);

        myth_read_shorts(plugin->tagfile, 13, iptr->unknown4);

        iptr->item_id2 = myth_read_short(plugin->tagfile);
        iptr->item_id3 = myth_read_short(plugin->tagfile);
      }
      
      bulkdata_size += mesh_header->num_items * sizeof(itempos);
    } else {
      itemarray = NULL;
    }
  }
}

// Free the big mesh arrays that we don't
// need any more (but keep the rest of the tag around).
void MeshTag::free_bulk_data(void)
{
  if (loaded_bulk_data) {
    loaded_bulk_data = FALSE;

    free(grid);
    if (itemtypearray)    free(itemtypearray);
    if (itemarray)        free(itemarray);
    if (scenery_lookup) free(scenery_lookup);

    grid = NULL;
    itemtypearray = NULL;
    itemarray = NULL;
    scenery_lookup = NULL;
  }
}

// Turn the mesh into a Bintree.
Bintree *MeshTag::make_bintree(int use_texture_borders)
{
  Bintree *bintree;
  X256Tag *colormap;
  TagHashStructure *hash;
  itemtype *itptr;
  itempos *iptr;
  int i, scenery_type, texture_index;
  vec3 item_location;

  // Read this mesh's parameters from the tag file.
  if (!loaded_data)
    read_data();

  // Load the bulk height array.
  load_bulk_data();

  // Set up the Bintree using this height data.
  bintree = new Bintree(this);

  // Get the colormap tag...
  hash = myth_tags->hash_structure(string_to_tagid(".256"));
  assert(hash != NULL);
  colormap = (X256Tag *) hash->retrieve_tag(mesh_header->colormap);

  if (colormap == NULL) {
    fprintf(debug_file, "WARNING: Cannot find colormap '%s' for the selected mesh!\n", tagid_to_string(mesh_header->colormap));
  } else {
    bintree->use_borders = use_texture_borders;
    colormap->setup_texturing(bintree);
  }

  if (draw_scenery) {
    // Get the scenery hash structure.
    hash = myth_tags->hash_structure(string_to_tagid("scen"));
    assert(hash != NULL);
  
    // Count all of the scenery textures (there will be one for every core entry attached
    // to every scenery type).
    num_scenery_textures = 0;
    for (i = 0, itptr = itemtypearray; i < mesh_header->num_itemtypes; i++, itptr++) {
      if ((itptr->type == ITEMTYPE_SCENERY) && draw_scenery_tag(itptr->item_tag)) {
        // Get the scenery tag.
        itptr->scenerytag = (ScenTag *) hash->retrieve_tag(itptr->item_tag);
        if (itptr->scenerytag == NULL) {
          fprintf(debug_file, "WARNING: mesh '%s' references scen '%s' but cannot find it!\n",
                  tagid_to_string(header->our_id), tagid_to_string(itptr->item_tag));
          continue;
        }
        itptr->scenerytag->read_data();

        // Get the base offset of the textures for this scenery type.
//        fprintf(debug_file, "itemtype %d is scenery, tex index %d\n", i, num_scenery_textures);
        itptr->bintree_scenery_index = num_scenery_textures;

//        fprintf(debug_file, "incrementing by %d (number of core types)\n",
//                itptr->scenerytag->num_core_entries());
        num_scenery_textures += itptr->scenerytag->num_core_entries();
      }
    }

    // Tell the Bintree to set up its arrays. (scenery textures were counted earlier
    // in our first pass through the itempos array)
//    fprintf(debug_file, "calling alloc_scenery, %d %d\n", num_scenery_textures, num_scenery_items);
//    fflush(debug_file);
    bintree->alloc_scenery(num_scenery_textures, num_scenery_items);

    // Set up the scenery's textures in the bintree.
    for (i = 0, itptr = itemtypearray; i < mesh_header->num_itemtypes; i++, itptr++)
      if (itptr->scenerytag != NULL)
        itptr->scenerytag->setup_bintree_textures(bintree, itptr->bintree_scenery_index);

    // Scenery textures have been set up and mipmapped. We can free all of the sprite collections.
    for (i = 0, itptr = itemtypearray; i < mesh_header->num_itemtypes; i++, itptr++)
      if (itptr->scenerytag != NULL)
        itptr->scenerytag->dump_sprite_data();

    for (i = 0, iptr = itemarray; i < mesh_header->num_items; i++, iptr++) {
      if (iptr->type != ITEMTYPE_SCENERY)
        continue;

      if ((iptr->itemtype_index < 0) || (iptr->itemtype_index >= num_scenery_types)) {
        fprintf(debug_file, "WARNING: item %d has itemtype_index %d, not in valid range [0 %d]\n",
                i, iptr->itemtype_index, num_scenery_types);
        continue;
      }

      scenery_type = scenery_lookup[iptr->itemtype_index];
      if ((scenery_type < 0) || (scenery_type >= mesh_header->num_itemtypes)) {
        fprintf(debug_file, "WARNING: item %d scenery_lookup[%d] = %d, not in valid range [0 %d]\n",
                i, iptr->itemtype_index, scenery_type, mesh_header->num_itemtypes);
        continue;
      }

      
      itptr = &itemtypearray[scenery_type];
      if (itptr->type != ITEMTYPE_SCENERY) {
        fprintf(debug_file, "WARNING: item %d looked up to itemtype %d, type %d... not scenery\n",
                i, scenery_type, itptr->type);
        continue;
      }

      // If this type of scenery isn't being drawn, continue.
      if (itptr->bintree_scenery_index == -1)
        continue;

      item_location.x = iptr->item_pos[0] / 512.0;
      item_location.y = iptr->item_pos[1] / 512.0;
      item_location.z = iptr->item_pos[2] - min_height;

      texture_index = itptr->bintree_scenery_index + (itptr->scenerytag->num_core_entries() * rand()) / RAND_MAX;
//      fprintf(debug_file, "texture index %d randomly out of %d-base %d-length\n",
//              texture_index, itptr->bintree_scenery_index, itptr->scenerytag->num_core_entries());

      bintree->create_scenery_object(texture_index, item_location);
    }
  }

  return bintree;
}

// Do we draw this kind of scenery?
int MeshTag::draw_scenery_tag(tag_id scenery_tag)
{
  // LMOTH flag.
  if (scenery_tag == string_to_tagid("nelf"))
    return 0;

  // Territories flags
  if (scenery_tag == string_to_tagid("netf"))
    return 0;

  // Stampede flags
  if (scenery_tag == string_to_tagid("nesf"))
    return 0;

  // KOTH flag
  if (scenery_tag == string_to_tagid("neko"))
    return 0;

  // BACON!!!
  if (scenery_tag == string_to_tagid("nest"))
    return 0;

  // Captures balls.
  if (scenery_tag == string_to_tagid("necb"))
    return 0;

  // Flag Rally flags.
  if ((scenery_tag == string_to_tagid("nef1")) ||
      (scenery_tag == string_to_tagid("nef2")) ||
      (scenery_tag == string_to_tagid("nef3")) ||
      (scenery_tag == string_to_tagid("nef4")) ||
      (scenery_tag == string_to_tagid("nef5")) ||
      (scenery_tag == string_to_tagid("nef6")) ||
      (scenery_tag == string_to_tagid("nef7")) ||
      (scenery_tag == string_to_tagid("nef8")) ||
      (scenery_tag == string_to_tagid("nef9")))
    return 0;

  // Scavenger Hunt balls.
  if ((scenery_tag == string_to_tagid("neb1")) ||
      (scenery_tag == string_to_tagid("neb2")) ||
      (scenery_tag == string_to_tagid("neb3")) ||
      (scenery_tag == string_to_tagid("neb4")) ||
      (scenery_tag == string_to_tagid("neb5")) ||
      (scenery_tag == string_to_tagid("neb6")) ||
      (scenery_tag == string_to_tagid("neb7")) ||
      (scenery_tag == string_to_tagid("neb8")) ||
      (scenery_tag == string_to_tagid("neb9")))
    return 0;

  // Balls on Parade
  if (scenery_tag == string_to_tagid("nebo"))
    return 0;

  // Capture the Flag
  if (scenery_tag == string_to_tagid("nect"))
    return 0;

  // Draw everything else.
//  fprintf(debug_file, "accepting scenery tag '%s'\n", tagid_to_string(scenery_tag));
  return 1;
}

