// MythPlugin.h : Definitions for class MythPlugin.
//

#define PLUGIN_NAME_LENGTH     31
#define PLUGIN_URL_LENGTH      63
#define ENTRYPOINT_NAME_LENGTH 63

#define MAX_PLUGIN_TAGS       20000           // largely arbitrary

#define PLUGINTYPE_TAG        0
#define PLUGINTYPE_PLUGIN     1
#define PLUGINTYPE_PATCH      2
#define PLUGINTYPE_FOUNDATION 3
#define PLUGINTYPE_CACHE      4
#define PLUGINTYPE_MAX        PLUGINTYPE_CACHE

// A plugin-header structure.
typedef struct pluginheader_ {
  short plugin_type;  // 0:tag, 1:plugin 2:patch 3:foundation 4:cache
  short version;      // only for 'patch' plugins

  char name[PLUGIN_NAME_LENGTH+1];  // Plugin name, null-terminated
  char url[PLUGIN_URL_LENGTH+1];    // Website URL, null-terminated

  short num_entrypoints;
  short num_tags;
  unsigned long checksum;
  unsigned long flags;
  long size;  // Entire file size, including this header

  unsigned long header_checksum;  // Checksum is calculated with this field zero

  long unused[1];

  tag_id plugin_signature;  // Unique plugin ID
} pluginheader;

// An entry-point structure. Every plugin file has one entrypoint
// structure for every tag of type 'mesh' that it contains.
typedef struct entrypoint_ {
  tag_id mesh;          // id of the mesh tag
  long unused[2];
  class MythPlugin *our_plugin;          // A pointer to our plugin. NB: It's just a guess
                                         // that this space is available for use, but there
                                         // never seems to be any info stored here in the tag files.
  char meshtagname[TAG_NAME_LENGTH+1];
  char meshname[ENTRYPOINT_NAME_LENGTH+1];
} entrypoint;

// This class encapsulates a Myth II plugin file.
class MythPlugin {
public:
  FILE *tagfile;
  pluginheader header;
  entrypoint *entrypoint_array;
  Tag **tags;

  // Is this plugin currently activated? Only relevant for plugins
  // of type PLUGINTYPE_PLUGIN... everything else is always activated.
  int plugin_activated;

  // Read the plugin from a file... called once the header has been
  // read and verified.
  MythPlugin(FILE *tagfile, pluginheader *copy_header);

  ~MythPlugin();
};

// Read a plugin from a file. First reads the plugin header, and then
// constructs the full plugin data type if it's valid.
MythPlugin *read_plugin(char *pluginname);
