// MythTag.h : Definitions for class MythTag and supporting code.
//

typedef long tag_id;
extern tag_id null_tag_id;

// Routines for reading Myth's platform-independent tag files.
unsigned char myth_read_byte(FILE *read_file);
short myth_read_short(FILE *read_file);
long myth_read_long(FILE *read_file);

inline tag_id myth_read_tag(FILE *read_file) {
  return myth_read_long(read_file);
}

void myth_read_bytes(FILE *read_file, int n_bytes, unsigned char *dest);
void myth_read_shorts(FILE *read_file, int n_shorts, short *dest);
void myth_read_longs(FILE *read_file, int n_longs, long *dest);

// Print out a tag-id (character by character). NB: uses a static buffer.
// Save the results if you want to keep them.
char *tagid_to_string(tag_id write_id);

// Convert a string into a tag-id.
tag_id string_to_tagid(char *tagstr);

#define TAG_NAME_LENGTH   31

// The tag-header data structure contains a single 'something' of data.
typedef struct tagheader_ {
  unsigned short datum_header;   // initialised at runtime, invalid on disk
  unsigned char flags;
  unsigned char type;

  char name[TAG_NAME_LENGTH+1];   // name of the tag

  tag_id tag_type;            // type of data, e.g. '.256' 'mesh' 'stli'
  tag_id our_id;             // ID of the particular tag within its type, user-defined

  long offset;               // offset of the tag's data with respect to the start of the
                             // file - no matter what kind of file (single tag, plugin, etc)
  long size;                 // bytes of data

  unsigned long userdata;    // these values are interpreted differently for each
  short version;             // type of tag

  char foundation_id;        // which monolithic file the tag will be put into when
                             // compacting tags for CD burning
                             //   small install       0
                             //   medium install      1
                             //   large install       2
                             //   small interface     3
                             //   large interface     4
  unsigned char priority;    // used for overwriting determination. invalid on disk

  tag_id telltale_signature; // always 'mth2'
} tagheader;

// Create a new tag from a header that we read from disk.
class Tag *new_tag(class MythPlugin *plugin, tagheader *header);

// A tag data file. Subclassed for every different kind of tag.
class Tag {
public:
  tagheader *header;

  class MythPlugin *plugin;
  int loaded_data;
  int loaded_bulk_data, bulkdata_size;

  // Read the minimal data from disk (constant-size structs, etc).
  virtual void read_data(void) = 0;

  // Read the bulk data from disk, or dump it after use.
  virtual void load_bulk_data(void) = 0;
  virtual void free_bulk_data(void) = 0;

  // Create a new tag.
  Tag(class MythPlugin *plugin, tagheader *header);

  // Delete the tag.
  ~Tag();
};

// A tag of an unimplemented type.
class UnimplementedTag : public Tag {
public:
  // Read the static data from disk.
  virtual void read_data(void);

  // Read any big arrays from disk, or dump them after use.
  virtual void load_bulk_data(void);
  virtual void free_bulk_data(void);

  // Create a new tag.
  UnimplementedTag(class MythPlugin *plugin, tagheader *header);

  // Delete the tag.
  ~UnimplementedTag();
};
