// PriQueue.h : Definitions for class PriQueue
//
// Manages a priority-queue for the Bintree; is either for splittable
// triangles or mergeable diamonds, ordered by either their priority or their
// recomputation distance.

const int queue_growsize = 1000;

// The information that a priority queue works on.
typedef struct priq_info_ {
  double priority;
  int priq_location, buf_index;
} priq_info;

class PriQueue {
public:  
  int num_elements, queue_bufsize, sort_order;
  priq_info **queue_buf;

  /*
   * Interface routines
   */

  // Create a new (blank) priority queue.
  PriQueue(int sort_order);

  // Clear the priority queue.
  void clear(void);

  // Insert an element into the priority queue.
  void insert_element(priq_info *element);

  // Return the highest-priority element in the queue.
  priq_info *get_element(void);

  // An element must be prematurely removed from the priority queue.
  void remove_element(priq_info *element);

  // Bubble an element up or down the heap from a starting location until
  // it reaches the correct location.
  void bubble_up(priq_info *element);
  void bubble_down(priq_info *element);
};
