// ScenTag.cpp : Routines for class ScenTag and supporting code.
//

#include "stdafx.h"


// Create a new tag.
ScenTag::ScenTag(MythPlugin *my_plugin, tagheader *header_struct)
  : Tag(my_plugin, header_struct)
{
  collection_reference = NULL;
  object_properties = NULL;
}

// Read the scen data from disk.
void ScenTag::read_data(void)
{
  // Seek the tag-file to the start of the tag's data.
  fseek(plugin->tagfile, header->offset, SEEK_SET);

  // Read the 128-byte scen data.
  scen_header.flags = myth_read_long(plugin->tagfile);
  scen_header.collection = myth_read_tag(plugin->tagfile);
  scen_header.object_id = myth_read_tag(plugin->tagfile);
  myth_read_longs(plugin->tagfile, 7, scen_header.unknown1);
  scen_header.collection_bitmapid = myth_read_short(plugin->tagfile);
  scen_header.unknown2 = myth_read_short(plugin->tagfile);
  myth_read_longs(plugin->tagfile, 21, scen_header.unknown3);

  loaded_data = TRUE;
}

// Read the bulk mesh data (the big arrays) from disk.
void ScenTag::load_bulk_data(void)
{
  // Do nothing.
  bulkdata_size = 0;
  loaded_bulk_data = TRUE;
}

// Free the big mesh arrays that we don't
// need any more (but keep the rest of the tag around).
void ScenTag::free_bulk_data(void)
{
  // Do nothing.
}

// Set up the scenery's texture in the bintree.
void ScenTag::setup_bintree_textures(Bintree *bintree, int texture_index)
{

  // Read this mesh's parameters from the tag file.
  if (!loaded_data)
    read_data();

  if ((collection_reference == NULL) || (object_properties == NULL))
    if (!load_tags())
      return;

  // Tell the collection reference to dump our sprite into the Bintree's texture.
  object_properties->read_data();
  collection_reference->setup_bintree_textures(bintree, texture_index,
                                               scen_header.collection_bitmapid,
                                               object_properties->obje_header.scale_factor);
}

// Get our defining tags.
int ScenTag::load_tags(void)
{
  TagHashStructure *hash;

  // Read the scenery tag.
  if (!loaded_data)
    read_data();

  // Get the object tag that defines our behaviour.
  hash = myth_tags->hash_structure(string_to_tagid("obje"));
  assert(hash != NULL);

  object_properties = (ObjeTag *) hash->retrieve_tag(scen_header.object_id);
  if (object_properties == NULL) {
    fprintf(debug_file, "ERROR: scenery item '%s' couldn't find object properties '%s'\n",
            tagid_to_string(header->our_id), tagid_to_string(scen_header.object_id));
    return FALSE;
  }

  // Get the collection reference.
  hash = myth_tags->hash_structure(string_to_tagid("core"));
  assert(hash != NULL);

  collection_reference = (CoreTag *) hash->retrieve_tag(scen_header.collection);
  if (collection_reference == NULL) {
    fprintf(debug_file, "ERROR: scenery item '%s' couldn't find collection reference '%s'\n",
            tagid_to_string(header->our_id), tagid_to_string(scen_header.collection));
    return FALSE;
  }

  return TRUE;
}

// How many core entries are there? Tells us how many textures we
// must allocate.
int ScenTag::num_core_entries(void)
{
  TagHashStructure *hash;

  if (collection_reference == NULL) {
    // Read the scenery tag.
    if (!loaded_data)
      read_data();

    // Get the collection reference.
    hash = myth_tags->hash_structure(string_to_tagid("core"));
    assert(hash != NULL);

    collection_reference = (CoreTag *) hash->retrieve_tag(scen_header.collection);
    if (collection_reference == NULL) {
      fprintf(debug_file, "ERROR: scenery item '%s' couldn't find collection reference '%s'\n",
              tagid_to_string(header->our_id), tagid_to_string(scen_header.collection));
      return 0;
    }
  }

  collection_reference->read_data();
  return collection_reference->data.num_entries;
}

// Dump our sprite data.
void ScenTag::dump_sprite_data(void)
{
  if (collection_reference != NULL)
    collection_reference->dump_sprite_data();
}

