// ScenTag.h : Definitions for class ScenTag and supporting code.
//

// The 128-byte header for the scen data.
typedef struct scenheader_ {
  long flags;           // unknown
  tag_id collection;    // collection ref  (of type 'core')
  tag_id object_id;     // object id       (of type 'obje')
  long unknown1[7];
  short collection_bitmapid; // id of bitmap within collection
  short unknown2;
  long unknown3[21];
} scenheader;

// A scen - a piece of scenery.
class ScenTag : public Tag {
public:
  scenheader scen_header;
  CoreTag *collection_reference;
  ObjeTag *object_properties;

  // Read the scen's static data from disk.
  virtual void read_data(void);

  // These virtual functions do nothing.
  virtual void load_bulk_data(void);
  virtual void free_bulk_data(void);

  // Set up the scenery's texture in the bintree.
  void setup_bintree_textures(class Bintree *bintree, int texture_index);

  // How many core entries are there? Tells us how many textures we
  // must allocate.
  int num_core_entries(void);

  // Get our defining tags.
  int load_tags(void);

  // Dump our sprite data.
  void dump_sprite_data(void);

  // Create a new tag.
  ScenTag(class MythPlugin *plugin, tagheader *header);

  // Delete the tag.
  ~ScenTag();
};
