// TagDatabase.h : Definitions for class TagDatabase.
//

const int pluginarray_growsize = 10;
const int hasharray_growsize = 10;

typedef struct taghashentry_ {
  // Info stored about this tag.
  tag_id our_id;
  void *tagdata;

  // Where this tag came from (used to determine precedence)
  MythPlugin *plugin;

  // The last iteration ID to return this tagentry (or another tagentry
  // with the same tag ID).
  int last_iterate_id;

  // Pointer to the next tag.
  struct taghashentry_ *next;
} taghashentry;

typedef struct iteratestruct_ {
  taghashentry *tagentry;
  int bucket_index;
  int iterate_id;
} iteratestruct;

class TagHashStructure {
public:
  tag_id tagtype;
  int num_tags, next_iterate_id;   // note: num_tags is the number of unique tag IDs,
                                   // there may be more hash entries than this.

  taghashentry *bucket[256];

  // Store a tag.
  void store_tag(tag_id store_id, void *tagdata, MythPlugin *owner);

  // Retrieve a particular tag.
  void *retrieve_tag(tag_id wanted_id, int ignore_plugin_activation = FALSE);

  // Iterate all tags in the structure, retrieving the highest-precedence one.
  // Just used for entrypoints usually.
  int iterate_tags(iteratestruct **iterator);

  // The hash function.
  unsigned char hashfunction(tag_id hash_id);

  // Create a new hash structure for a given kind of tag.
  TagHashStructure(tag_id tagtype);
  ~TagHashStructure(void);
};

class TagDatabase {
public:
  int num_plugins, pluginarraysize;
  MythPlugin **pluginarray;

  // Hash structure for every kind of tag, and one for entrypoints.
  int num_hashes, hasharraysize;
  TagHashStructure **hasharray;
  TagHashStructure *entrypoint_hash;

  // Read all the plugins in a given directory.
  void read_plugins(char *directory);

  // Register a plugin.
  void register_plugin(MythPlugin *plugin);

  // Get the hash structure for a particular kind of tag.
  TagHashStructure *hash_structure(long tagtype);

  // Create an empty tag database, and delete one.
  TagDatabase(void);
  ~TagDatabase(void);
};
