// targa.cpp : The TGA writing routine.
//

#include "stdafx.h"

// bytes_per_pixel is 1 for greyscale, 3 for RGB
void write_tga(char *filename, char bytes_per_pixel,
               short img_x, short img_y, unsigned char *data)
{
  FILE *out_file;
  unsigned char *imgptr;
  int i;

  if ((out_file = fopen(filename, "wb")) == NULL) {
    fprintf(debug_file, "ERROR: write_tga couldn't open file '%s'\n", filename);
    return;
  }

  fputc(0, out_file);    // length of image identification field (not included)
  fputc(0, out_file);    // no colormap included

  // This image is TGA type 3 for uncompressed greyscale, or 2 for uncompressed RGB
  fputc((bytes_per_pixel == 1) ? 3 : 2, out_file);

  // Color-map specification (just zeroes);
  fputc(0, out_file);
  fputc(0, out_file);
  fputc(0, out_file);
  fputc(0, out_file);
  fputc(0, out_file);

  // X and Y origin of the image (0, 0). 2 bytes each
  fputc(0, out_file);
  fputc(0, out_file);
  fputc(0, out_file);
  fputc(0, out_file);

  // Width and height of the image.
  write_tga_short(out_file, img_x);
  write_tga_short(out_file, img_y);

  // Bits per pixel
  fputc(8 * bytes_per_pixel, out_file);

  // Flag byte - bits are:
  //
  // 0-3     number of attribute bits           0
  // 4       reserved                           0
  // 5       screen-origin bit, 0 = lower left  0
  //                            1 = upper left
  // 6-7     storage interleaving flag          0
  fputc(0, out_file);

  // Write the image data.
  if (bytes_per_pixel == 1) {
    fwrite((void *) data, img_x * img_y, 1, out_file);
  } else {
    // TGA files are BGR so we must reverse each pixel's data.
    for (i = 0, imgptr = data; i < img_x * img_y; i++) {
      fputc(imgptr[2], out_file);
      fputc(imgptr[1], out_file);
      fputc(imgptr[0], out_file);

      imgptr += 3;
    }
  }

  fclose(out_file);
}

// Write a TGA-format short integer (little-endian format)
void write_tga_short(FILE *out_file, short write_int)
{
  fputc(write_int & 0xFF, out_file);
  write_int >>= 8;
  fputc(write_int & 0xFF, out_file);
}
