/*
* vec3.h
* vec3 stuff for particle-system.
*/

#ifndef MIN
#define MIN(a, b) ((a) < (b) ? (a) : (b))
#endif

#ifndef MAX
#define MAX(a, b) ((a) < (b) ? (b) : (a))
#endif

#define SQR(x)      ((x) * (x))

class vec3 {
public:
  double x, y, z;

  inline vec3() {
  }

  inline vec3(double ax, double ay, double az)
  {
    x = ax; y = ay; z = az;
  }

  inline void zero(void)
  {
    x = 0; y = 0; z = 0;
  }

  inline friend double mag(vec3 &v)
  {
    return sqrt(v * v);
  }

  inline void operator= (double *a)
  {
    x = a[0]; y = a[1]; z = a[2];
  }

  inline void operator= (double a)
  {
    x = a; y = a; z = a;
  }

  inline friend vec3 operator+ (vec3 &a, vec3 &b)
  {
    return vec3(a.x + b.x, a.y + b.y, a.z + b.z);
  }
  
  inline vec3 &operator+= (vec3 &a)
  {
    x += a.x;
    y += a.y;
    z += a.z;
    
    return *this;
  }
  
  inline friend vec3 operator- (vec3 &a, vec3 &b)
  {
    return vec3(a.x - b.x, a.y - b.y, a.z - b.z);
  }
  
  inline friend vec3 operator- (vec3 &v)
  {
    return vec3(-v.x, -v.y, -v.z);
  }

  inline vec3 &operator-= (vec3 &a)
  {
    x -= a.x;
    y -= a.y;
    z -= a.z;
    
    return *this;
  }
  
  inline friend vec3 operator* (double s, vec3 &v)
  {
    return vec3(v.x * s, v.y * s, v.z * s);
  }
  
  inline friend vec3 operator* (vec3 &v, double s)
  {
    return vec3(v.x * s, v.y * s, v.z * s);
  }
  
  inline vec3 &operator*= (double s)
  {
    x *= s;
    y *= s;
    z *= s;
    
    return *this;
  }
  
  inline friend vec3 operator/ (vec3 &v, double s)
  {
    return vec3(v.x / s, v.y / s, v.z / s);
  }
  
  inline vec3 &operator/= (double s)
  {
    x /= s;
    y /= s;
    z /= s;
    
    return *this;
  }

  // Dot product
  inline friend double operator* (vec3 &a, vec3 &b)
  {
    return (a.x * b.x + a.y * b.y + a.z * b.z);
  }
  
  // Cross product
  inline friend vec3 operator^ (vec3 &a, vec3 &b)
  {
    return vec3(a.y * b.z - a.z * b.y, a.z * b.x - a.x * b.z, a.x * b.y - a.y * b.x);
  }
  
  inline friend void normalize(vec3 &v)
  {
    v /= mag(v);
  }

  inline void read(double *array)
  {
    x = array[0]; y = array[1]; z = array[2];
  }

  inline void store(double *array)
  {
    array[0] = x; array[1] = y; array[2] = z;
  }

  inline friend vec3 vmin(vec3 &a, vec3 &b)
  {
    return vec3(MIN(a.x, b.x), MIN(a.y, b.y), MIN(a.z, b.z));
  }

  inline friend vec3 vmax(vec3 &a, vec3 &b)
  {
    return vec3(MAX(a.x, b.x), MAX(a.y, b.y), MAX(a.z, b.z));
  }
};







