// X256Tag.cpp : Routines for class X256Tag and supporting code.
//

#include "stdafx.h"

#define INTERPOLATE_HUECHANGE    FALSE
#define WRITE_SCENERY_TGA        FALSE

// Create a new tag.
X256Tag::X256Tag(MythPlugin *my_plugin, tagheader *header_struct)
  : Tag(my_plugin, header_struct)
{
}

// Read the .256 data from disk.
void X256Tag::read_data(void)
{
  // Seek the tag-file to the start of the tag's data.
  fseek(plugin->tagfile, header->offset, SEEK_SET);

  // Read the 320-byte .256 header.
  x256_header = (x256header *) malloc(sizeof(x256header));
  if (!x256_header) mem_fail(sizeof(x256header),
                             "X256Tag header");

//  fprintf(debug_file, "reading .256 '%s' from %s @ %d\n", tagid_to_string(header->our_id),
//          plugin->header.name, header->offset);
  
  x256_header->unknown1 = myth_read_long(plugin->tagfile);
  x256_header->type = myth_read_long(plugin->tagfile);
  x256_header->unknown2 = myth_read_long(plugin->tagfile);
  x256_header->unknown3 = myth_read_long(plugin->tagfile);
  myth_read_longs(plugin->tagfile, 12, x256_header->unknown4);

  x256_header->num_palettes = myth_read_long(plugin->tagfile);
  x256_header->palette_offset = myth_read_long(plugin->tagfile);
  x256_header->palette_length = myth_read_long(plugin->tagfile);
  x256_header->unknown10 = myth_read_long(plugin->tagfile);
//  fprintf(debug_file, "%d palettes at %d (len %d)\n",
//          x256_header->num_palettes, x256_header->palette_offset, x256_header->palette_length);

  x256_header->num_huechanges = myth_read_long(plugin->tagfile);
  x256_header->huechange_offset = myth_read_long(plugin->tagfile);
  x256_header->huechange_length = myth_read_long(plugin->tagfile);
  x256_header->unknown14 = myth_read_long(plugin->tagfile);
//  fprintf(debug_file, "%d huechanges at %d (len %d)\n",
//          x256_header->num_huechanges, x256_header->huechange_offset, x256_header->huechange_length);

  x256_header->num_sectioninfo = myth_read_long(plugin->tagfile);
  x256_header->sectioninfo_offset = myth_read_long(plugin->tagfile);
  x256_header->sectioninfo_length = myth_read_long(plugin->tagfile);
  x256_header->unknown15 = myth_read_long(plugin->tagfile);

  x256_header->num_bitmapdescriptors = myth_read_long(plugin->tagfile);
  x256_header->bitmapdescriptor_offset = myth_read_long(plugin->tagfile);
  x256_header->bitmapdescriptor_length = myth_read_long(plugin->tagfile);
  x256_header->unknown16 = myth_read_long(plugin->tagfile);

  x256_header->num_bitmapsets = myth_read_long(plugin->tagfile);
  x256_header->bitmapset_offset = myth_read_long(plugin->tagfile);
  x256_header->bitmapset_length = myth_read_long(plugin->tagfile);
  x256_header->unknown17 = myth_read_long(plugin->tagfile);

  x256_header->num_others = myth_read_long(plugin->tagfile);
  x256_header->other_offset = myth_read_long(plugin->tagfile);
  x256_header->other_length = myth_read_long(plugin->tagfile);
  x256_header->unknown21 = myth_read_long(plugin->tagfile);

  x256_header->num_colormaps = myth_read_long(plugin->tagfile);
  x256_header->colormap_offset = myth_read_long(plugin->tagfile);
  x256_header->colormap_length = myth_read_long(plugin->tagfile);
  x256_header->unknown22 = myth_read_long(plugin->tagfile);

  x256_header->num_colormapindices = myth_read_long(plugin->tagfile);
  x256_header->colormapindex_offset = myth_read_long(plugin->tagfile);
  x256_header->colormapindex_length = myth_read_long(plugin->tagfile);
  x256_header->unknown23 = myth_read_long(plugin->tagfile);

  x256_header->unknown24 = myth_read_long(plugin->tagfile);
  x256_header->unknown25 = myth_read_long(plugin->tagfile);
  x256_header->unknown26 = myth_read_long(plugin->tagfile);
  x256_header->unknown27 = myth_read_long(plugin->tagfile);

  x256_header->unknown28 = myth_read_long(plugin->tagfile);
  x256_header->unknown29 = myth_read_long(plugin->tagfile);
  x256_header->unknown30 = myth_read_long(plugin->tagfile);
  x256_header->unknown31 = myth_read_long(plugin->tagfile);

  x256_header->unknown32 = myth_read_long(plugin->tagfile);
  x256_header->unknown33 = myth_read_long(plugin->tagfile);
  x256_header->unknown34 = myth_read_long(plugin->tagfile);
  x256_header->unknown35 = myth_read_long(plugin->tagfile);

  x256_header->unknown36 = myth_read_long(plugin->tagfile);
  x256_header->unknown37 = myth_read_long(plugin->tagfile);
  x256_header->bulk_offset = myth_read_long(plugin->tagfile);
  x256_header->bulk_length = myth_read_long(plugin->tagfile);

  myth_read_longs(plugin->tagfile, 16, x256_header->unknown38);


  if (x256_header->bulk_offset != sizeof(x256header)) {
    fprintf(debug_file, "ERROR reading .256 '%s' - header corrupt (bulk_offset %d != sizeof(header) %d)\n",
            tagid_to_string(header->our_id), x256_header->bulk_offset, sizeof(x256header));
    premature_shutdown();
  }

  loaded_data = TRUE;
}

// Read the bulk .256 data (the big graphics arrays) from disk.
void X256Tag::load_bulk_data(void)
{
  int i, j, read_bytes, palettedatasize;
  paletteinfo *paletteptr;
  huechange *hueptr;

  if (!loaded_bulk_data) {
    loaded_bulk_data = TRUE;
    bulkdata_size = 0;

    if (x256_header->num_palettes > 0) {
      // Allocate memory for the palette array.
      palettearray = (paletteinfo *) malloc(x256_header->num_palettes
                                            * sizeof(paletteinfo));
      if (!palettearray) mem_fail(x256_header->num_palettes * sizeof(paletteinfo),
                                 "X256Tag palette array");
      bulkdata_size += x256_header->num_palettes * sizeof(paletteinfo);

      // Seek to the correct location in the tagfile.
      fseek(plugin->tagfile, header->offset + x256_header->bulk_offset + x256_header->palette_offset, SEEK_SET);

      for (i = 0, paletteptr = palettearray; i < x256_header->num_palettes; i++, paletteptr++) {
        paletteptr->num_colors = myth_read_long(plugin->tagfile);
        myth_read_longs(plugin->tagfile, 6, paletteptr->unknown);

        // Read (and discard) the long which we use to store the colorval pointer in.
        myth_read_long(plugin->tagfile);

        // Allocate space for the palette data.
        palettedatasize = 8 * paletteptr->num_colors;
        paletteptr->colordata = (unsigned char *) malloc(palettedatasize * sizeof(unsigned char));
        if (!paletteptr->colordata) mem_fail(palettedatasize * sizeof(unsigned char),
                                             "X256Tag palette data");
        bulkdata_size += palettedatasize;

        read_bytes = fread((void *) paletteptr->colordata, sizeof(unsigned char),
                           palettedatasize, plugin->tagfile);
        if (read_bytes != palettedatasize) {
          fprintf(debug_file, "ERROR: Premature end-of-file reading .256 '%s' palette %d colordata - read_bytes %d != size %d\n",
                  tagid_to_string(header->our_id), i, read_bytes, palettedatasize);
          premature_shutdown();
        }
      }
    } else {
      palettearray = NULL;
    }

    if (x256_header->num_huechanges > 0) {
      // Allocate memory for the huechange array.
      huechangearray = (huechange *) malloc(x256_header->num_huechanges
                                            * sizeof(huechange));
      if (!huechangearray) mem_fail(x256_header->num_huechanges * sizeof(huechange),
                                    "X256Tag hue change array");
      bulkdata_size += x256_header->num_huechanges * sizeof(huechange);

      // Seek to the correct location in the tagfile.
      fseek(plugin->tagfile, header->offset + x256_header->bulk_offset +
            x256_header->huechange_offset, SEEK_SET);

//      fprintf(debug_file, "  reading %d huechanges\n", x256_header->num_huechanges);
      for (i = 0, hueptr = huechangearray; i < x256_header->num_huechanges; i++, hueptr++) {
        myth_read_bytes(plugin->tagfile, 64, (unsigned char *) hueptr->name);
//        fprintf(debug_file, "    huechange %d: %s\n", i, hueptr->name);

        hueptr->hue_min = myth_read_short(plugin->tagfile);
        hueptr->hue_max = myth_read_short(plugin->tagfile);
//        fprintf(debug_file, "    hue range %d - %d (%f - %f)\n",
//                hueptr->hue_min, hueptr->hue_max,
//                360 * (hueptr->hue_min / 65535.0), 360 * (hueptr->hue_max / 65535.0));

        myth_read_shorts(plugin->tagfile, 17, hueptr->unknown2);
//        for (j = 0; j < 17; j++)
//          fprintf(debug_file, "    unknown2[%d] 0x%04X\n", j, hueptr->unknown2[j]);

        hueptr->hue_median = myth_read_short(plugin->tagfile);
//        fprintf(debug_file, "    hue median %d (%f)\n",
//                hueptr->hue_median, 360 * (hueptr->hue_median / 65535.0));

        myth_read_longs(plugin->tagfile, 6, hueptr->unknown3);
//        for (j = 0; j < 6; j++)
//          fprintf(debug_file, "    unknown3[%d] 0x%08X\n", j, hueptr->unknown3[j]);
      }
    } else {
      huechangearray = NULL;
    }

    switch(x256_header->type) {
    case X256TYPE_COLORMAP: {
      sectioninfo *sectionptr;
      bitmapinfo *bitmapptr;
      int imagedatasize;

      // Allocate memory for the sectioninfo array.
      sectioninfoarray = (sectioninfo *) malloc(x256_header->num_sectioninfo
                                                * sizeof(sectioninfo));
      if (!sectioninfoarray) mem_fail(x256_header->num_sectioninfo * sizeof(sectioninfo),
                                      "X256Tag sectioninfo array");
      bulkdata_size += x256_header->num_sectioninfo * sizeof(sectioninfo);
      
      // Seek to the correct location in the tagfile.
      fseek(plugin->tagfile, header->offset + x256_header->bulk_offset
            + x256_header->sectioninfo_offset, SEEK_SET);
      
      for (i = 0, sectionptr = sectioninfoarray;
           i < x256_header->num_sectioninfo; i++, sectionptr++) {

        myth_read_longs(plugin->tagfile, 16, sectionptr->unknown1);        
        sectionptr->imagedata_offset = myth_read_long(plugin->tagfile);
        sectionptr->imagedata_length = myth_read_long(plugin->tagfile);        
        sectionptr->unknown2 = myth_read_long(plugin->tagfile);
        sectionptr->unknown3 = myth_read_long(plugin->tagfile);
        myth_read_longs(plugin->tagfile, 12, sectionptr->unknown4);
      }
      
      // Allocate memory for the bitmapinfo array.
      bitmaparray = (bitmapinfo *) malloc(x256_header->num_sectioninfo
                                          * sizeof(bitmapinfo));
      if (!bitmaparray) mem_fail(x256_header->num_sectioninfo * sizeof(sectioninfo),
                                 "X256Tag bitmap array");
      bulkdata_size += x256_header->num_sectioninfo * sizeof(bitmapinfo);
      
      for (i = 0, bitmapptr = bitmaparray;
           i < x256_header->num_sectioninfo; i++, bitmapptr++) {
        // Seek to the correct location in the tagfile.
        fseek(plugin->tagfile, header->offset + x256_header->bulk_offset
          + sectioninfoarray[i].imagedata_offset, SEEK_SET);

        myth_read_shorts(plugin->tagfile, 3, bitmapptr->unknown1);
        bitmapptr->unknown2 = myth_read_short(plugin->tagfile);
        myth_read_shorts(plugin->tagfile, 3, bitmapptr->unknown3);
        myth_read_shorts(plugin->tagfile, 7, bitmapptr->unknown4);
        
        bitmapptr->unknown5[0] = myth_read_short(plugin->tagfile);
        bitmapptr->unknown5[1] = myth_read_short(plugin->tagfile);
        
        bitmapptr->img_x = myth_read_short(plugin->tagfile);
        bitmapptr->img_y = myth_read_short(plugin->tagfile);

        myth_read_shorts(plugin->tagfile, 6, bitmapptr->unknown6);
        
        // Read (and discard) the long which we use to store the image pointer in.
        myth_read_long(plugin->tagfile);
        
        // Allocate space for the image data.
        imagedatasize = bitmapptr->img_x * bitmapptr->img_y;
        bitmapptr->imagedata = (unsigned char *) malloc(imagedatasize * sizeof(unsigned char));
        if (!bitmapptr->imagedata) mem_fail(imagedatasize * sizeof(unsigned char),
                                            "X256Tag bitmap data");
        bulkdata_size += imagedatasize;
        
        read_bytes = fread((void *) bitmapptr->imagedata, sizeof(unsigned char),
          imagedatasize, plugin->tagfile);
        if (read_bytes != imagedatasize) {
          fprintf(debug_file, "ERROR: Premature end-of-file reading .256 '%s' bitmap %d imagedata - read_bytes %d != size %d\n",
            tagid_to_string(header->our_id), i, read_bytes, imagedatasize);
          premature_shutdown();
        }
      }
      break;
    }
    default: {
      // We assume any .256 that we don't know about is a sprite
      // collection. Chances are that we're right.
      spriteinfo *spriteptr;
      spriteheader sprite_header;
      bitmapdescriptor *bitdescptr;
      bitmapset *bitsetptr;
      viewelement *elemptr;
      long *rowinfo;
      short segmentinfo[SPRITE_MAX_SEGMENTS * 2];
      int row_index, segment_index, num_segments, num_pixels;
      int pixel_start_pos, run_start, run_length;

      // Allocate memory for the sectioninfo array.
      spriteinfoarray = (spriteinfo *) malloc(x256_header->num_sectioninfo
                                              * sizeof(spriteinfo));
      if (!spriteinfoarray) mem_fail(x256_header->num_sectioninfo * sizeof(spriteinfo),
                                     "X256Tag spriteinfo array");
      bulkdata_size += x256_header->num_sectioninfo * sizeof(spriteinfo);
      
      // Seek to the correct location in the tagfile.
      fseek(plugin->tagfile, header->offset + x256_header->bulk_offset
            + x256_header->sectioninfo_offset, SEEK_SET);
      
      for (i = 0, spriteptr = spriteinfoarray;
           i < x256_header->num_sectioninfo; i++, spriteptr++) {
      
        myth_read_bytes(plugin->tagfile, sizeof(spriteptr->name), (unsigned char *) spriteptr->name);
        myth_read_longs(plugin->tagfile, 8, spriteptr->unknown1);
        
        spriteptr->sprite_offset = myth_read_long(plugin->tagfile);
        spriteptr->sprite_length = myth_read_long(plugin->tagfile);
        
        spriteptr->unknown2 = myth_read_long(plugin->tagfile);
        
        spriteptr->width  = myth_read_short(plugin->tagfile);
        spriteptr->height = myth_read_short(plugin->tagfile);
        myth_read_longs(plugin->tagfile, 11, spriteptr->unknown3);

        myth_read_long(plugin->tagfile);
        spriteptr->spritedata = NULL;
      }

      // Seek to the start of the first sprite.
      fseek(plugin->tagfile, header->offset + x256_header->bulk_offset +
            spriteinfoarray[0].sprite_offset, SEEK_SET);

      // Read every sprite.
      for (i = 0, spriteptr = spriteinfoarray; i < x256_header->num_sectioninfo; i++, spriteptr++) {
        // Jump to the start of the sprite.
        fseek(plugin->tagfile, header->offset +
               x256_header->bulk_offset + spriteptr->sprite_offset, SEEK_SET);

        // Read the 52-byte sprite header.
        sprite_header.width  = myth_read_short(plugin->tagfile);
        sprite_header.height = myth_read_short(plugin->tagfile);
        assert(sprite_header.width  == spriteptr->width);
        assert(sprite_header.height == spriteptr->height);

        sprite_header.rawdata_rowlen = myth_read_short(plugin->tagfile);
        sprite_header.unknown2 = myth_read_short(plugin->tagfile);
        sprite_header.unknown3 = myth_read_short(plugin->tagfile);
        myth_read_shorts(plugin->tagfile, 19, sprite_header.unknown4);

        // Allocate memory for this sprite. Initialise it to palette entry zero.
        spriteptr->spritedata = (unsigned char *) calloc(spriteptr->width * spriteptr->height,
                                                         sizeof(unsigned char));
        if (!spriteptr->spritedata) mem_fail(spriteptr->width * spriteptr->height * sizeof(unsigned char),
                                             "X256Tag sprite data");
        bulkdata_size += spriteptr->width * spriteptr->height;

        // Read the rowInfo long for every row.
        rowinfo = (long *) malloc(sprite_header.height * sizeof(long));
        if (!rowinfo) mem_fail(sprite_header.height * sizeof(long),
                               "X256Tag sprite rowinfo");
        myth_read_longs(plugin->tagfile, sprite_header.height, rowinfo);

        if (sprite_header.rawdata_rowlen == -1) {
          // Segment-based compressed sprite data.
          for (row_index = 0; row_index < sprite_header.height; row_index++) {
            num_segments = myth_read_short(plugin->tagfile);
            num_pixels   = myth_read_short(plugin->tagfile);
            if ((num_segments >= SPRITE_MAX_SEGMENTS) || (num_pixels > sprite_header.width)) {
              // There is something wrong - we have bungled the sprite reading.
              // Set this sprite to transparent and go to the next one.
              memset(spriteptr->spritedata, 0, spriteptr->width * spriteptr->height);
              break;
            }
            assert(num_segments < SPRITE_MAX_SEGMENTS);
            assert(num_pixels <= sprite_header.width);

            // Read the segment data.
            myth_read_shorts(plugin->tagfile, 2 * num_segments, segmentinfo);

            pixel_start_pos = ftell(plugin->tagfile);
            for (segment_index = 0; segment_index < num_segments; segment_index++) {
              run_start = segmentinfo[2 * segment_index];
              run_length = segmentinfo[2 * segment_index + 1] - run_start;
              assert((run_length > 0) && (run_start + run_length <= sprite_header.width));
              myth_read_bytes(plugin->tagfile, run_length,
                              spriteptr->spritedata + row_index * sprite_header.width + run_start);
            }
            assert(ftell(plugin->tagfile) == pixel_start_pos + num_pixels);

            // Make certain that we short-align ourselves if an odd number of pixels.
            if (num_pixels % 2)
              myth_read_byte(plugin->tagfile);          
          }
        } else {
          assert(sprite_header.rawdata_rowlen == sprite_header.width);
          // Read the raw data in uncompressed format.
          num_pixels = sprite_header.rawdata_rowlen * sprite_header.height;
          myth_read_bytes(plugin->tagfile, num_pixels, spriteptr->spritedata);

          // Make certain that we short-align ourselves if an odd number of pixels.
          if (num_pixels % 2)
            myth_read_byte(plugin->tagfile);          
        }
        free(rowinfo);
      }

      if (x256_header->num_bitmapdescriptors > 0) {
        // Allocate memory for the bitmapdescriptor array.
        bitmapdescriptorarray = (bitmapdescriptor *) malloc(x256_header->num_bitmapdescriptors
                                                            * sizeof(bitmapdescriptor));
        if (!bitmapdescriptorarray) mem_fail(x256_header->num_bitmapdescriptors * sizeof(bitmapdescriptor),
                                             "X256Tag bitmapdescriptor array");
        bulkdata_size += x256_header->num_bitmapdescriptors * sizeof(bitmapdescriptor);
      
        // Seek to the correct location in the tagfile.
        fseek(plugin->tagfile, header->offset + x256_header->bulk_offset
              + x256_header->bitmapdescriptor_offset, SEEK_SET);

        for (i = 0, bitdescptr = bitmapdescriptorarray;
             i < x256_header->num_bitmapdescriptors; i++, bitdescptr++) {
          // Read the 64-byte bitmapdescriptor.
          myth_read_shorts(plugin->tagfile, 6, bitdescptr->unknown1);
        
          bitdescptr->origin_x = myth_read_short(plugin->tagfile);
          bitdescptr->origin_y = myth_read_short(plugin->tagfile);
          myth_read_shorts(plugin->tagfile, 6, bitdescptr->unknown2);

          bitdescptr->sprite_index = myth_read_short(plugin->tagfile);

          myth_read_shorts(plugin->tagfile, 17, bitdescptr->unknown3);
        }
      } else {
        bitmapdescriptorarray = NULL;
      }
      
      if (x256_header->num_bitmapsets > 0) {
        // Allocate memory for the bitmapset array.
        bitmapsetarray = (bitmapset *) malloc(x256_header->num_bitmapsets
                                            * sizeof(bitmapset));
        if (!bitmapsetarray) mem_fail(x256_header->num_bitmapsets * sizeof(bitmapset),
                                     "X256Tag bitmapset array");
        bulkdata_size += x256_header->num_bitmapsets * sizeof(bitmapset);
      
        // Seek to the correct location in the tagfile.
        fseek(plugin->tagfile, header->offset + x256_header->bulk_offset
              + x256_header->bitmapset_offset, SEEK_SET);

        for (i = 0, bitsetptr = bitmapsetarray;
             i < x256_header->num_bitmapsets; i++, bitsetptr++) {
          // Read the 128-byte bitmapset.
          myth_read_bytes(plugin->tagfile, 64, (unsigned char *) bitsetptr->name);
        
          bitsetptr->viewdata_offset = myth_read_long(plugin->tagfile);
          bitsetptr->viewdata_length = myth_read_long(plugin->tagfile);
 
          bitsetptr->unknown1 = myth_read_short(plugin->tagfile);
          bitsetptr->unknown2 = myth_read_short(plugin->tagfile);

          myth_read_longs(plugin->tagfile, 13, bitsetptr->unknown3);
        }

        // Read each bitmapset's viewset header and list of view elements.
        for (i = 0, bitsetptr = bitmapsetarray; i < x256_header->num_bitmapsets; i++, bitsetptr++) {
          // Jump to the offset of the viewset.
          fseek(plugin->tagfile, header->offset + x256_header->bulk_offset
                + bitsetptr->viewdata_offset, SEEK_SET);

          myth_read_shorts(plugin->tagfile, 5, bitsetptr->viewset.unknown1);

          bitsetptr->viewset.num_elements = myth_read_short(plugin->tagfile);
          assert((bitsetptr->viewset.num_elements > 0) && (bitsetptr->viewset.num_elements < 1000));

          // heh, I understand so little of this structure that it's scary
          myth_read_shorts(plugin->tagfile, 26, bitsetptr->viewset.unknown2);

          // Allocate the view-element list for this bitmapset.
          bitsetptr->element = (viewelement *) malloc(bitsetptr->viewset.num_elements *
                                                      sizeof(viewelement));
          if (!bitsetptr->element) mem_fail(bitsetptr->viewset.num_elements * sizeof(viewelement),
                                            "X256Tag bitmapset viewelement array");
          bulkdata_size += bitsetptr->viewset.num_elements * sizeof(viewelement);

          for (j = 0, elemptr = bitsetptr->element;
               j < bitsetptr->viewset.num_elements; j++, elemptr++) {
            // Read the viewelement.
            elemptr->unknown1 = myth_read_short(plugin->tagfile);

            myth_read_shorts(plugin->tagfile, 22, elemptr->unknown2);

            elemptr->bitmapdesc_index = myth_read_short(plugin->tagfile);
            assert((elemptr->bitmapdesc_index >= 0) &&
                   (elemptr->bitmapdesc_index < x256_header->num_bitmapdescriptors));
          }
        }
      } else {
        bitmapsetarray = NULL;
      }

      break;
    }
      fprintf(debug_file, "ERROR: Unknown .256 type %d!\n", x256_header->type);
      premature_shutdown();
    }
  }
}

// Free the big mesh arrays that we don't
// need any more (but keep the rest of the tag around).
void X256Tag::free_bulk_data(void)
{
  int i;

  if (loaded_bulk_data) {
    loaded_bulk_data = FALSE;

    for (i = 0; i < x256_header->num_palettes; i++)
      free(palettearray[i].colordata);
    if (palettearray) free(palettearray);
    palettearray = NULL;

    if (x256_header->type == X256TYPE_COLORMAP) {
      // Free the colormap structures.
      for (i = 0; i < x256_header->num_sectioninfo; i++)
        if (bitmaparray[i].imagedata)
          free(bitmaparray[i].imagedata);

      if (bitmaparray) free(bitmaparray);
      bitmaparray = NULL;

      if (sectioninfoarray) free(sectioninfoarray);
      sectioninfoarray = NULL;

    } else {
      // We assume anything that isn't a colormap is a sprite collection.
      // After all, anything that isn't one of these two classes
      // won't be loaded by the app.

      // Free the bitmap sets and view elements, etc.
      for (i = 0; i < x256_header->num_bitmapsets; i++)
        if (bitmapsetarray[i].element)
          free(bitmapsetarray[i].element);

      if (bitmapsetarray) free(bitmapsetarray);
      bitmapsetarray = NULL;

      if (bitmapdescriptorarray) free(bitmapdescriptorarray);
      bitmapdescriptorarray = NULL;

      // Free the sprite structures.
      for (i = 0; i < x256_header->num_sectioninfo; i++)
        if (spriteinfoarray[i].spritedata)
          free(spriteinfoarray[i].spritedata);

      if (spriteinfoarray) free(spriteinfoarray);
      spriteinfoarray = NULL;
    }
  }
}


// Set up the texturing parameters for a Bintree.
void X256Tag::setup_texturing(Bintree *bintree)
{
  int i, j, num_sections;
  unsigned char *unified_texture, *shadowptr, *colorptr, *textureptr;
  unsigned char shadow_amount, color_index;

  // Read this .256's parameters from the tag file.
  if (!loaded_data)
    read_data();

  // Load the bulk graphics arrays.
  if (!loaded_bulk_data)
    load_bulk_data();

  // Set up the Bintree using the graphics data.
  // There must be one texture and one shadow bitmap for every
  // section of the mesh that the Bintree indexes.
  num_sections = bintree->x_sections * bintree->y_sections;
  if (x256_header->num_sectioninfo != 2 * num_sections) {
    fprintf(debug_file, "ERROR: mesh texturemap .256 '%s' has %d bitmaps, isn't 2 * mesh sections (%d)\n",
            tagid_to_string(header->our_id), x256_header->num_sectioninfo, num_sections);
    premature_shutdown();
  }

  // Allocate space for a unified-texture array. Will be reused for every section.
  unified_texture = (unsigned char *) malloc(3 * SQR(256) * sizeof(unsigned char));
  if (!unified_texture) mem_fail(3 * SQR(256) * sizeof(unsigned char),
                                 "X256Tag unified colormap / shadowmap");

  bintree->init_texture(num_sections, 256);
  for (i = 0; i < num_sections; i++) {
    if ((bitmaparray[2*i  ].img_x != 256) || (bitmaparray[2*i  ].img_y != 256)) {
      fprintf(debug_file, "ERROR: mesh texturemap .256 '%s' colormap %d size %dx%d != 256x256\n",
              tagid_to_string(header->our_id), 2*i,
              bitmaparray[2*i].img_x, bitmaparray[2*i].img_y);
      premature_shutdown();
    }
    if ((bitmaparray[2*i+1].img_x != 256) || (bitmaparray[2*i+1].img_y != 256)) {
      fprintf(debug_file, "ERROR: mesh texturemap .256 '%s' shadowmap %d size %dx%d != 256x256\n",
              tagid_to_string(header->our_id), 2*i + 1,
              bitmaparray[2*i+1].img_x, bitmaparray[2*i+1].img_y);
      premature_shutdown();
    }

    // Put the shadowmap and the colormap together to form a unified texturemap.

    colorptr = bitmaparray[2*i].imagedata;
    shadowptr = bitmaparray[2*i + 1].imagedata;
    textureptr = unified_texture;

    for (j = 0; j < SQR(256); j++) {
      // The colormap value indexes the first palette to give us RGB values
      // which we multiply by the shadowmap value.
      shadow_amount = *shadowptr++;
      color_index = *colorptr++;

      *textureptr++ = (shadow_amount * palettearray[0].colordata[8 * color_index + 2*0]) / 255;
      *textureptr++ = (shadow_amount * palettearray[0].colordata[8 * color_index + 2*1]) / 255;
      *textureptr++ = (shadow_amount * palettearray[0].colordata[8 * color_index + 2*2]) / 255;
    }

    // Store this texture in the Bintree.
    bintree->build_miplevels(i, unified_texture);
  }
  bintree->textures_done();

  // Free the unified texture.
  free(unified_texture);

  // We no longer need our bulk graphics arrays, and can free them.
  free_bulk_data();
}

// X256TYPE_SPRITE: set up a scenery object's texture in the bintree.
void X256Tag::bintree_sprite_texture(Bintree *bintree, int texture_index,
                                     int bitmap_id, short scale_factor,
                                     int tintval, unsigned char *tintdata,
                                     coreentry *coreptr)
{
  int bitdesc_index, sprite_index, i, j, x, y, w, h, x_offset;
  spriteinfo *sprite;
  unsigned char *rgba_texture, *texptr, *spriteptr, *paletteptr;
  unsigned char palette_index, *newpalette, *newpaletteptr;
  unsigned long *longtexptr;

  // The tint color, in HSV, and interpolation fraction
  int tint_hue;
  double tintcol[3], tint_interp;

  // The colors that we're changing to, in HSV (0 - 1)
  double color[3][3];

  // The hue ranges that are being changed.
  int num_ranges;
  double hue_min[3], hue_max[3], hue_median[3], falloff_low[3], falloff_hi[3];

  double our_h, our_s, our_v, hue_offset, interp;
#if WRITE_SCENERY_TGA
  unsigned char *tgadata, *tgaptr;

  // Print this out so we know which texture is which tga.
  fprintf(debug_file, "*** X256Tag::bintree_sprite_texture, texture_index %d\n", texture_index);
#endif
  

  // Read this .256's parameters from the tag file, and the bulk graphics arrays.
  if (!loaded_data)
    read_data();
  if (!loaded_bulk_data)
    load_bulk_data();

  /*
   * Build the modified palette
   */

  tint_interp = tintval / 256.0;
  rgb_to_hsv(tintdata[0], tintdata[2], tintdata[4],
             &tintcol[0], &tintcol[1], &tintcol[2]);
  if (tintcol[1] < 1e-03) {
    // The tint does not change the hue.
    tint_hue = FALSE;
  } else {
    tint_hue = TRUE;
  }
//  fprintf(debug_file, "tinting by %f to %f %f %f (hue = %s)\n",
//          tint_interp, tintcol[0], tintcol[1], tintcol[2], tint_hue ? "TRUE" : "FALSE");

  for (j = 0; j < 3; j++)
    rgb_to_hsv(coreptr->colors[j][0], coreptr->colors[j][2], coreptr->colors[j][4],
               &color[j][0], &color[j][1], &color[j][2]);

  num_ranges = x256_header->num_huechanges;
  if (num_ranges == 0) {
    // There are no huechanges specifically set up in this file.
    // Can't be sure, but I *think* that in this case there is meant
    // to be a default huechange of the bright-green.
    num_ranges = 1;
    hue_min[0]     = 1.9 / 6.0;
    hue_median[0]  = 2.0 / 6.0;
    hue_max[0]     = 2.1 / 6.0;
    falloff_low[0] = 0.1 / 6.0;
    falloff_hi[0]  = 0.1 / 6.0;
  } else {
    for (j = 0; j < num_ranges; j++) {
      hue_min[j]    = huechangearray[j].hue_min    / 65535.0;
      hue_max[j]    = huechangearray[j].hue_max    / 65535.0;
      hue_median[j] = huechangearray[j].hue_median / 65535.0;

      falloff_low[j] = fabs(hue_median[j] - hue_min[j]);
      if (falloff_low[j] > 0.5)
        falloff_low[j] = 1.0 - falloff_low[j];

      falloff_hi[j]  = fabs(hue_median[j] - hue_max[j]);
      if (falloff_hi[j] > 0.5)
        falloff_hi[j]  = 1.0 - falloff_hi[j];
    }
  }

  newpalette = (unsigned char *) malloc(4 * 256);
  if (!newpalette) mem_fail(4 * 256,
                            "X256Tag huechange new palette");

  paletteptr = palettearray[0].colordata;
  newpaletteptr = newpalette;
  for (i = 0; i < 256; i++) {
    // We must check to see whether this palette entry falls within any of the
    // hue ranges that are being changed.
//    fprintf(debug_file, "palette %03d(%02X): %02X %02X %02X %02X (flags %02X %02X %02X %02X)\n", i, i,
//      paletteptr[0], paletteptr[2], paletteptr[4], paletteptr[6],
//      paletteptr[1], paletteptr[3], paletteptr[5], paletteptr[7]);
      
    rgb_to_hsv(paletteptr[0], paletteptr[2], paletteptr[4],
               &our_h, &our_s, &our_v);
      
//    fprintf(debug_file, "  our hsv %f %f %f\n", our_h, our_s, our_v);

    for (j = 0; j < num_ranges; j++) {
//      fprintf(debug_file, "  check range %d:\n", j);

      hue_offset = our_h - hue_median[j];
      if (hue_offset > 0.5)
        hue_offset -= 1.0;
      else if (hue_offset < -0.5)
        hue_offset += 1.0;
//      fprintf(debug_file, "    location relative to median: %f\n", hue_offset);
      
      if (hue_offset > 0) {
        // This color is above the median. Is it within the range?
        interp =  hue_offset / falloff_hi[j];
      } else {
        interp = -hue_offset / falloff_low[j];
      }
//      fprintf(debug_file, "  interp: %f\n", interp);
      
      if (interp < 1) {
        // We are within the range. Interpolate to the desired color.
//        fprintf(debug_file, "  interpolate to %f %f %f\n",
//                color[j][0], color[j][1], color[j][2]);
#if INTERPOLATE_HUECHANGE
        our_h = interp * our_h + (1 - interp) * color[j][0];
        our_s *= interp + (1 - interp) * color[j][1];
        our_v *= interp + (1 - interp) * color[j][2];
#else
        our_h  = color[j][0];
        our_s *= color[j][1];
        our_v *= color[j][2];
#endif
//        fprintf(debug_file, "  new color %f %f %f\n", our_h, our_s, our_v);

        // Don't bother checking other hue ranges, we have altered the color.
        break;
      }
    }

    // Apply the tint in HSV space.
    if (tintval > 0) {
      if (tint_hue)
        our_h = (1 - tint_interp) * our_h + tint_interp * tintcol[0];

      our_s *= (1 - tint_interp) + tint_interp * tintcol[1];
      our_v *= (1 - tint_interp) + tint_interp * tintcol[2];
//      fprintf(debug_file, "tinted by (%f of %f %f %f, %c) to %f %f %f\n",
//              tint_interp, tintcol[0], tintcol[1], tintcol[2], (tint_hue) ? 'T' : 'F',
//              our_h, our_s, our_v);
    }
    
    hsv_to_rgb(our_h, our_s, our_v, newpaletteptr);
//    fprintf(debug_file, "  RGB is %d %d %d\n",
//            newpaletteptr[0], newpaletteptr[1], newpaletteptr[2]);
    newpaletteptr[3] = paletteptr[7];  // alpha is unchanged

    newpaletteptr += 4;
    paletteptr += 8;
  }

  // Find the bitmap that matches the name provided.
  if ((bitmap_id < 0) || (bitmap_id >= x256_header->num_bitmapsets)) {
    fprintf(debug_file, "ERROR: scenery tried to access .256 collection '%s' sprite '%d' outside our range [0 %d]!\n",
            tagid_to_string(header->our_id), bitmap_id, 0, x256_header->num_bitmapsets);
    return;
  }

  // Get the first view element and extract the frame's bitmapdescriptor index.
  assert((bitmapsetarray[bitmap_id].viewset.num_elements > 0) &&
         (bitmapsetarray[bitmap_id].element != NULL));
  bitdesc_index = bitmapsetarray[bitmap_id].element[0].bitmapdesc_index;

  // Get the sprite index that this references.
  assert((bitdesc_index >= 0) && (bitdesc_index < x256_header->num_bitmapdescriptors));
  sprite_index = bitmapdescriptorarray[bitdesc_index].sprite_index;

  // Finally, get the sprite.
  assert((sprite_index >= 0) && (sprite_index < x256_header->num_sectioninfo));
  sprite = &spriteinfoarray[sprite_index];

  assert((sprite->width  > 0) && (sprite->width  <= 256) &&
         (sprite->height > 0) && (sprite->height <= 256));

  // Calculate the smallest powers of two greater than or equal to the sprite's width and height.
  for (w = 1; w < sprite->width; w <<= 1)
    ;
  for (h = 1; h < sprite->height; h <<= 1)
    ;

  // Work out where the sprite is mapped into this texture (it is always mapped to zero
  // in y, because this is where it touches the ground).
  x_offset = (w - sprite->width) / 2;

  // Create the texture and initialise it to empty (RGB = 0, alpha = 1).
  rgba_texture = (unsigned char *) malloc(4 * w * h);
  if (!rgba_texture) mem_fail(4 * w * h,
                              "X256Tag scenery texture");
  for (i = 0, longtexptr = (unsigned long *) rgba_texture; i < w * h; i++)
    *longtexptr++ = 0xFF000000;  // little endian, PC

  spriteptr = sprite->spritedata;

#if WRITE_SCENERY_TGA
  tgadata = (unsigned char *) malloc(3 * sprite->height * sprite->width);
  tgaptr = tgadata;
#endif

  for (y = 0; y < sprite->height; y++) {
    texptr = &rgba_texture[4 * (x_offset + (sprite->height - y) * w)];
    for (x = 0; x < sprite->width; x++) {
      palette_index = *spriteptr++;
      if (palette_index == 0) {
#if WRITE_SCENERY_TGA
        *tgaptr++ = 0x00;
        *tgaptr++ = 0x00;
        *tgaptr++ = 0xFF;
#endif
        texptr += 4; // leave as transparent
      } else {
        *texptr++ = newpalette[4*palette_index + 0];
        *texptr++ = newpalette[4*palette_index + 1];
        *texptr++ = newpalette[4*palette_index + 2];
        *texptr++ = newpalette[4*palette_index + 3];
#if WRITE_SCENERY_TGA
        *tgaptr++ = newpalette[4*palette_index + 0];
        *tgaptr++ = newpalette[4*palette_index + 1];
        *tgaptr++ = newpalette[4*palette_index + 2];
#endif
      }
    }
  }

#if WRITE_SCENERY_TGA
  char filename[256];
  sprintf(filename, "scenery\\tex%02d.tga", texture_index);
  write_tga(filename, 3, sprite->width, sprite->height, tgadata);
  free(tgadata);
#endif

  bintree->scenery_texture(texture_index, scale_factor, w, h, rgba_texture);

  free(newpalette);
}

// HSV to RGB and vice-versa
void rgb_to_hsv(unsigned char  r, unsigned char  g, unsigned char b,
                double *h, double *s, double *v)
{
  unsigned char maxval;
  double hue, delta;

  maxval = MAX(r, MAX(g, b));
  if (maxval == 0) {
    *h = *s = *v = 0;
    return;
  } else {
    *v = maxval / 255.0;
  }
  delta = maxval - MIN(r, MIN(g, b));
  if (delta == 0) {
    *h = *s = 0;
    return;
  }

  *s = delta / maxval;
  if (r == maxval) {
    if (g > b)
      hue = 0 + (g - b) / delta;
    else
      hue = 6 + (g - b) / delta;
  } else if (g == maxval) {
    hue = 2 + (b - r) / delta;
  } else {
    hue = 4 + (r - g) / delta;
  }

  *h = hue / 6;

//  fprintf(debug_file, "rgb_to_hsv on %d %d %d -> %f %f %f\n",
//          r, g, b, *h, *s, *v);
}

#define SET_RGBVAL(i, a)  returncol[i] = (unsigned char) (255 * a + 0.5)
#define SET_RGB(a, b, c) {  \
  SET_RGBVAL(0, a);         \
  SET_RGBVAL(1, b);         \
  SET_RGBVAL(2, c);         \
}

void hsv_to_rgb(double h, double s, double v, unsigned char *returncol)
{
  int sextant;
  double interp, p, q, t;

  if (s < 1e-03) {
    returncol[0] = returncol[1] = returncol[2] = 0;
    return;
  }

  h *= 6;
  sextant = (int) h;
  if (sextant >= 6)
    sextant -= 6;
  interp = h - sextant;

  p = v * (1 - s);
  q = v * (1 - (s *      interp));
  t = v * (1 - (s * (1 - interp)));

  switch(sextant) {
  case 0:
    SET_RGB(v, t, p);
    break;
  case 1:
    SET_RGB(q, v, p);
    break;
  case 2:
    SET_RGB(p, v, t);
    break;
  case 3:
    SET_RGB(p, q, v);
    break;
  case 4:
    SET_RGB(t, p, v);
    break;
  case 5:
    SET_RGB(v, p, q);
    break;
  }

//  fprintf(debug_file, "hsv_to_rgb on %f %f %f -> %d %d %d\n",
//          h, s, v, returncol[0], returncol[1], returncol[2]);
}
