// X256Tag.h : Definitions for class X256Tag and supporting code.
//

#define X256TYPE_COLORMAP   256

#define SPRITE_MAX_SEGMENTS 20

// The 320-byte header for the .256 data.
typedef struct x256header_ {
  long unknown1;
  long type;                     // seems to be 256 for colormap, 32 for overhead, 12 for sprite
  long unknown2;
  long unknown3;

  long unknown4[12];             // filler?

  long num_palettes;             // number of 2080-byte palettes
  long palette_offset;           // offset of palette (from bulk_offset)
  long palette_length;           // length of palette
  long unknown10;

  long num_huechanges;           // number of 128-byte hue-change sections (1 or 2)
  long huechange_offset;         // offset of huechange (from bulk_offset)
  long huechange_length;         // length of huechange
  long unknown14;

  long num_sectioninfo;          // number of 128-byte sectioninfo structs
  long sectioninfo_offset;       // offset of sectioninfo (from bulk_offset)
  long sectioninfo_length;       // length of sectioninfo
  long unknown15;

  long num_bitmapdescriptors;    // number of 64-byte bitmapdescriptors
  long bitmapdescriptor_offset;  // offset of bitmapdescriptor (from bulk_offset)
  long bitmapdescriptor_length;  // length of bitmapdescriptor
  long unknown16;

  long num_bitmapsets;           // number of 128-byte bitmapsets
  long bitmapset_offset;         // offset of bitmapset (from bulk_offset)
  long bitmapset_length;         // length of bitmapset
  long unknown17;

  long num_others;               // number of items in yet another array
  long other_offset;             // array offset
  long other_length;             // array length
  long unknown21;

  long num_colormaps;            // number of 65536-byte colormaps
  long colormap_offset;          // offset of colormap (from bulk_offset)
  long colormap_length;          // length of colormap
  long unknown22;

  long num_colormapindices;      // number of 256-byte colormapindices
  long colormapindex_offset;     // offset of colormapindex (from bulk_offset)
  long colormapindex_length;     // length of colormapindex
  long unknown23;

  long unknown24;                // another array?
  long unknown25;
  long unknown26;
  long unknown27;

  long unknown28;                // another array?
  long unknown29;
  long unknown30;
  long unknown31;

  long unknown32;                // another array?
  long unknown33;
  long unknown34;
  long unknown35;

  long unknown36;
  long unknown37;
  long bulk_offset;              // from the start of the tag's info
  long bulk_length;              // length of the bulk data arrays

  long unknown38[16];            // filler?
} x256header;

// The sectioninfo structure stores an entry in the sectionInfo table.
typedef struct sectioninfo_ {
  long unknown1[16];             // filler?

  long imagedata_offset;         // offset of imagedata (from bulk_offset)
  long imagedata_length;         // length of imagedata
  long unknown2;
  long unknown3;

  long unknown4[12];             // filler?
} sectioninfo;

// A 52-byte bitmap header in the .256 collection.
typedef struct bitmapinfo_ {
  short unknown1[3];             // appear to be maxval in R, G, B?
  short unknown2;                // rumored to be related to type of bitmap
  short unknown3[3];             // appear to be number of bits in R, G, B?
  short unknown4[7];             // usually zero
  short unknown5[2];             // number of miplevels maybe? appear to be log_2 of x and y res
  short img_x, img_y;            // dimensions of bitmap
  short unknown6[6];
  unsigned char *imagedata;      // nb: there appears to be random garbage at this location
                                 // in the tag files. I'm guessing it's a pointer location in memory
                                 // and isn't some horribly critical info. :)
} bitmapinfo;

// A palette in the .256 collection.
typedef struct paletteinfo_ {
  long num_colors;               // each color is 4 values: r g b a. each is represented by a byte for color
                                 // value and a byte for flags. flags are zero for most terrain, appear to be
                                 // 0x7F, 0xFF, 0x80 for things like blood (?).
  long unknown[6];
  unsigned char *colordata;      // nb: there appears to be random garbage at this location
                                 // in the tag files. I'm guessing it's a pointer location in memory
                                 // and isn't some horribly critical info. :)
} paletteinfo;

// A 128-byte huechange structure that represents a section of the
// palette that can be fiddled with by a 'core' tag.
typedef struct huechange_ {
  char name[64];
  unsigned short hue_min, hue_max;// huerange, min and max from 0 - 65535
  short unknown2[17];
  unsigned short hue_median;     // median of huerange
  long unknown3[6];
} huechange;

// A 128-byte sprite info structure in the .256 collection.
typedef struct spriteinfo_ {
  char name[32];                 // name of the sprite
  long unknown1[8];
  long sprite_offset;            // offset of the sprite data (from bulk_offset)
  long sprite_length;            // length of the sprite data
  long unknown2;                 
  short width;                   // width in pixels
  short height;                  // height in pixels
  long unknown3[11];             // filler?
  unsigned char *spritedata;     // overwrite whatever's stored here (hopefully not important)
} spriteinfo;

// A sprite header in the .256 collection.
typedef struct spriteheader_ {
  short width;                   // width in pixels
  short height;                  // height in pixels
  short rawdata_rowlen;          // -1 for segment-encoded, any other indicates raw data with rowlen bytes / row
  short unknown2;
  short unknown3;
  short unknown4[19];            // filler?
} spriteheader;

// A 64-byte bitmapdescriptor that references a sprite.
typedef struct bitmapdescriptor_ {
  short unknown1[6];
  short origin_x, origin_y;      // coordinates of the object's origin, measured from top left (y = h - orig_y)
  short unknown2[6];
  short sprite_index;            // index into the spriteinfo array
  short unknown3[17];
} bitmapdescriptor;

// A 64-byte header for a number of view-elements that make up a set of views.
typedef struct viewsetheader_ {
  short unknown1[5];
  short num_elements;            // number of 48-byte view-elements following this header
  short unknown2[26];
} viewsetheader;

// A 48-byte element that indexes a bitmapdescriptor for a frame
typedef struct viewelement_ {
  short unknown1;                // sometimes the same as bitmapdesc_index, sometimes -1
  short unknown2[22];            // zero for scenery
  short bitmapdesc_index;        // index into the bitmapdescriptor array
} viewelement;

// A 128-byte bitmapset that binds a number of bitmapdescriptors together
// into an object with (possibly) multiple views, animation sequences, etc. The Big Kahuna.
typedef struct bitmapset_ {
  char name[64];
  long viewdata_offset;          // relative to bulk_offset
  long viewdata_length;          // length of viewdata
  short unknown1;                // appears to be constant 013D... flags?
  short unknown2;                // varies wildly
  long unknown3[13];             // filler?

  viewsetheader viewset;         // header for the list of view elements
  viewelement *element;          // the list of view elements
} bitmapset;

// A .256 - a collection of graphics.
class X256Tag : public Tag {
public:
  x256header *x256_header;

  /*
   * Bulk arrays
   */
  paletteinfo *palettearray;
  huechange *huechangearray;

  // X256TYPE_COLORMAP
  sectioninfo *sectioninfoarray;
  bitmapinfo *bitmaparray;

  // X256TYPE_SPRITE
  spriteinfo *spriteinfoarray;
  bitmapdescriptor *bitmapdescriptorarray;
  bitmapset *bitmapsetarray;

  // Read the .256's static data from disk.
  virtual void read_data(void);

  // Read the big graphics arrays from disk, or dump them after use.
  virtual void load_bulk_data(void);
  virtual void free_bulk_data(void);

  // Create a new tag.
  X256Tag(class MythPlugin *plugin, tagheader *header);

  // X256TYPE_COLORMAP: set up the colormap texturing parameters for a Bintree.
  void setup_texturing(Bintree *bintree);

  // X256TYPE_SPRITE: set up a scenery object's texture in the bintree.
  void bintree_sprite_texture(Bintree *bintree, int texture_index,
                              int bitmap_id, short scale_factor,
                              int tintval, unsigned char *tintcol,
                              coreentry *coreptr);

  // Delete the tag.
  ~X256Tag();
};

void rgb_to_hsv(unsigned char  r, unsigned char  g, unsigned char b,
                double *h, double *s, double *v);
void hsv_to_rgb(double h, double s, double v, unsigned char *returncol);
